module bandpass_mod
  ! Module containing various types of filters
  !	- gaussian
  !	- hann
  !
  ! Jason Chang
  ! April 2014
  ! update (11/13/14): if alpha is zero, use standard dev. of 10% central

  !==========================================================
  ! includes
  !==========================================================
  use fourier1d_mod
  implicit none

  !==========================================================
  ! declare module variables
  !==========================================================
  integer,private                           :: n1,nf,opt
  real,private                              :: of,df
  real,private,parameter                    :: pi=3.1415926535
  real,private,allocatable,dimension (:)    :: freq,filter

  contains


  !==========================================================
  ! bandpass initialization subroutine
  !==========================================================
  subroutine bandpass_init(n1_in,o1_in,d1_in,opt_in)

    !===========================================
    ! declare variables
    !===========================================
    integer                    :: ii
    integer                    :: n1_in,opt_in
    real                       :: o1_in,d1_in

    !===========================================
    ! pass variables into entire modules
    !===========================================
    n1 = n1_in
    opt = opt_in

    !===========================================
    ! allocate arrays
    !===========================================
    allocate(filter(n1)); filter=0
    allocate(freq(n1_in)); freq=0

    !===========================================
    ! calculate fourier units
    !===========================================
    nf = n1_in
    df = 1/((n1_in-1)*d1_in)
    of = -1./(2*d1_in)
    do ii=1,n1
      freq(ii) = of+(ii-1)*df
    end do

    !===========================================
    ! initialize fft
    !===========================================
    call fft1d_init(n1)
    
  end subroutine


  !==========================================================
  ! close gaussian subroutine
  !==========================================================
  subroutine bandpass_close()

    !===========================================
    ! deallocate input and output arrays
    !===========================================
    deallocate(freq,filter)

    !===========================================
    ! close fft stuff
    !===========================================
    call fft1d_close()

  end subroutine


  !==========================================================
  ! gaussian filter subroutine
  !==========================================================
  subroutine apply_bandpass(cdat,cfreq,alpha)

    !=============================
    ! declare variables
    !=============================
    integer                           :: ii
    integer                           :: f1_ind,f2_ind
    real                              :: f1,f2
    real                              :: cfreq,alpha,arg
    complex,dimension (:)             :: cdat

    !=============================
    ! take fft
    !=============================
    call fft1d(.false.,1.,cdat)

    !============================================
    ! construct filter
    !============================================
    select case(opt)

      !==============================
      ! heaviside
      !==============================
      case(0)
        do ii=1,n1/2
          cdat(ii) = 0
        end do
        do ii=n1/2+2,n1
          cdat(ii) = 2.*cdat(ii)
        end do

      !==============================
      ! gaussian
      !==============================
      case(1)
      if (alpha.ne.0) then
        do ii=1,n1
          arg = (freq(ii)-cfreq)/cfreq
          filter(ii) = exp(-alpha*(arg**2))
        end do
      else
        do ii=1,n1
          arg = freq(ii)-cfreq
          filter(ii) = exp(-(arg**2)/(2*(0.1*cfreq)**2))
        end do
      end if
      cdat = cdat*filter

      !==============================
      ! gaussian+heaviside
      !==============================
      case(2)
        ! multiply by modified heaviside function
        do ii=1,n1/2
          cdat(ii) = 0
        end do
        do ii=n1/2+2,n1
          cdat(ii) = 2.*cdat(ii)
        end do

        ! apply gaussian filter
        if (alpha.ne.0) then
          do ii=1,n1
            arg = (freq(ii)-cfreq)/cfreq
            filter(ii) = exp(-alpha*(arg**2))
          end do
        else
          do ii=1,n1
            arg = freq(ii)-cfreq
            filter(ii) = exp(-(arg**2)/(2*((0.1*cfreq)**2)))
          end do
        end if
        cdat = cdat*filter


      !==============================
      ! hann
      !==============================
      case(3)
        ! get frequency indices
        f1 = cfreq-alpha
        f2 = cfreq+alpha
        f1_ind = nint((f1-of)/df)+1
        f2_ind = nint((f2-of)/df)+1

        ! apply hann filter
        do ii=f1_ind,f2_ind
          filter(ii)= .5*(1.-cos(2.*pi*(ii-f1_ind)/real(f2_ind-f1_ind)))
        end do
        cdat = cdat*filter

    end select

    !=============================
    ! take ifft
    !=============================
    call fft1d(.true.,1.,cdat)

  end subroutine



  !==========================================================
  ! gaussian filter subroutine
  !==========================================================
  subroutine gaussian_filter(cdat,cfreq,alpha)

    !=============================
    ! declare variables
    !=============================
    integer                           :: ii
    real                              :: cfreq,alpha,arg
    complex,dimension (:)             :: cdat

    !=============================
    ! take fft
    !=============================
    call fft1d(.false.,1.,cdat)

    !=============================
    ! apply filter
    !=============================
    do ii=1,n1
      arg = (freq(ii)-cfreq)/cfreq
      filter(ii) = exp(-alpha*(arg**2))
    end do
    cdat = cdat*filter

    !=============================
    ! take ifft
    !=============================
    call fft1d(.true.,1.,cdat)

  end subroutine



  !==========================================================
  ! gaussian filter with heaviside subroutine
  !==========================================================
  subroutine gaussian_heaviside_filter(cdat,cfreq,alpha)

    !=============================
    ! declare variables
    !=============================
    integer                           :: ii
    real                              :: cfreq,alpha,arg
    complex,dimension (:)             :: cdat

    !=============================
    ! take fft
    !=============================
    call fft1d(.false.,1.,cdat)

    !=============================
    ! apply filter
    !=============================
    ! multiply by modified heaviside function
    do ii=1,n1/2
      cdat(ii) = 0
    end do
    do ii=n1/2+2,n1
      cdat(ii) = 2.*cdat(ii)
    end do

    ! apply gaussian filter
    if (alpha.ne.0) then
      do ii=1,n1
        arg = (freq(ii)-cfreq)/cfreq
        filter(ii) = exp(-alpha*(arg**2))
      end do
    else
      do ii=1,n1
        arg = freq(ii)-cfreq
        filter(ii) = exp(-(arg**2)/(2*(0.1*cfreq)**2))
      end do
    end if
    cdat = cdat*filter

    !=============================
    ! take ifft
    !=============================
    call fft1d(.true.,1.,cdat)

  end subroutine



  !==========================================================
  ! heaviside filter subroutine (mostly for analytic signal)
  !==========================================================
  subroutine heaviside_filter(cdat)

    !=============================
    ! declare variables
    !=============================
    integer                           :: ii
    complex,dimension (:)             :: cdat

    !=============================
    ! take fft
    !=============================
    call fft1d(.false.,1.,cdat)

    !=============================
    ! apply filter
    !=============================
    ! multiply by modified heaviside function
    do ii=1,n1/2
      cdat(ii) = 0
    end do
    do ii=n1/2+2,n1
      cdat(ii) = 2.*cdat(ii)
    end do

    !=============================
    ! take ifft
    !=============================
    call fft1d(.true.,1.,cdat)

  end subroutine
  


  !==========================================================
  ! hann filter
  !==========================================================
  subroutine hann_filter(cdat,f1,f2)

    !=============================
    ! declare variables
    !=============================
    integer                           :: ii
    integer                           :: f1_ind,f2_ind
    real                              :: f1,f2
    real,parameter                    :: pi=3.1415926535
    complex,dimension (:)             :: cdat

    !=============================
    ! get frequency indices
    !=============================
    f1_ind = nint((f1-of)/df)+1
    f2_ind = nint((f2-of)/df)+1

    !=============================
    ! take fft
    !=============================
    call fft1d(.false.,1.,cdat)

    !=============================
    ! apply filter
    !=============================
    do ii=f1_ind,f2_ind
      filter(ii)= .5*(1.-cos(2.*pi*(ii-f1_ind)/real(f2_ind-f1_ind)))
    end do
    cdat = cdat*filter

    !=============================
    ! take ifft
    !=============================
    call fft1d(.true.,1.,cdat)

  end subroutine

end module
