module fourier1d_mod
  ! Module containing fourier transforms and inverse fourier transforms
  ! Created to perform 2d FFTs
  !
  ! February 2014
  ! Jason Chang
  
  !===========================================
  ! standard declarations
  !===========================================
  implicit none
  include "fftw3.f"


  !===========================================
  ! declare module variables
  !===========================================
  integer,private                           :: n1
  complex,allocatable,dimension (:),private :: in1,out1
  integer*8,private                         :: fft_for,fft_inv

  contains

  !=============================================================
  ! initialization subroutine
  !=============================================================
  subroutine fft1d_init(n1_in)

    !===========================================
    ! declare variables
    !===========================================
    integer :: n1_in

    !===========================================
    ! set these variables for entire module
    !===========================================
    n1 = n1_in

    !===========================================
    ! allocate input and output arrays
    !===========================================
    allocate(in1(n1)); in1=0.
    allocate(out1(n1)); out1=0.

    !===========================================
    ! set up fft plans
    !===========================================
    call sfftw_plan_dft_1d(fft_for,n1,in1,out1,FFTW_FORWARD,FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(fft_inv,n1,out1,in1,FFTW_BACKWARD,FFTW_ESTIMATE)

  end subroutine




  !=============================================================
  ! close subroutine
  !=============================================================
  subroutine fft1d_close()

    !===========================================
    ! deallocate input and output arrays
    !===========================================
    deallocate(in1,out1)

    !===========================================
    ! destroy fft plans
    !===========================================
    call sfftw_destroy_plan(fft_for)
    call sfftw_destroy_plan(fft_inv)

  end subroutine




  !=============================================================
  ! fft on first axis
  !=============================================================
  subroutine fft1d(adj,sign1,cdata)

    !===========================================
    ! input local variables
    !===========================================
    integer                             :: ii
    complex,dimension (:)               :: cdata
    real                                :: sign1
    logical,intent(in)                  :: adj


    !===========================================
    ! kernels for fft and ifft
    !===========================================
    ! if adjoint, then do ifft
    if (adj) then
      ! perform ifftshift
      out1(1:(n1+1)/2) = cdata(n1-(n1+1)/2+1:n1)
      out1((n1+1)/2+1:n1) = cdata(1:n1-(n1+1)/2)
      ! perform ifft
      call sfftw_execute(fft_inv)
      ! write out to cdata
      cdata = in1
    else
      ! grab data
      in1 = cdata
      ! perform fft
      call sfftw_execute(fft_for)
      ! perform fftshift
      cdata(n1-(n1+1)/2+1:n1) = out1(1:(n1+1)/2)
      cdata(1:n1-(n1+1)/2) = out1((n1+1)/2+1:n1)
    end if

    ! scale cdata
    cdata = sign1*cdata/sqrt(real(n1))

  end subroutine


end module
