module fourier_mod
  ! Module containing fourier transforms and inverse fourier transforms
  ! Created to perform 2d FFTs
  !
  ! February 2014
  ! Jason Chang
  
  !===========================================
  ! standard declarations
  !===========================================
  implicit none
  include "fftw3.f"


  !===========================================
  ! declare module variables
  !===========================================
  integer,private                           :: n1,n2
  complex,allocatable,dimension (:),private :: in1,out1,in2,out2
  integer*8,private                         :: fft_for1,fft_inv1,fft_for2,fft_inv2

  contains

  !=============================================================
  ! initialization subroutine
  !=============================================================
  subroutine fft_init(n1_in,n2_in)

    !===========================================
    ! declare variables
    !===========================================
    integer :: n1_in,n2_in

    !===========================================
    ! set these variables for entire module
    !===========================================
    n1 = n1_in
    n2 = n2_in

    !===========================================
    ! allocate input and output arrays
    !===========================================
    allocate(in1(n1)); in1=0.
    allocate(in2(n2)); in2=0.
    allocate(out1(n1)); out1=0.
    allocate(out2(n2)); out2=0.

    !===========================================
    ! set up fft plans
    !===========================================
    call sfftw_plan_dft_1d(fft_for1,n1,in1,out1,FFTW_FORWARD,FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(fft_inv1,n1,out1,in1,FFTW_BACKWARD,FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(fft_for2,n2,in2,out2,FFTW_FORWARD,FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(fft_inv2,n2,out2,in2,FFTW_BACKWARD,FFTW_ESTIMATE)

  end subroutine




  !=============================================================
  ! close subroutine
  !=============================================================
  subroutine fft_close()

    !===========================================
    ! deallocate input and output arrays
    !===========================================
    deallocate(in1,out1,in2,out2)

    !===========================================
    ! destroy fft plans
    !===========================================
    call sfftw_destroy_plan(fft_for1)
    call sfftw_destroy_plan(fft_inv1)
    call sfftw_destroy_plan(fft_for2)
    call sfftw_destroy_plan(fft_inv2)

  end subroutine




  !=============================================================
  ! fft on first axis
  !=============================================================
  subroutine fft_axis1(adj,sign1,cdata)

    !===========================================
    ! input local variables
    !===========================================
    integer                             :: ii
    complex,allocatable,dimension (:,:) :: cdata
    real                                :: sign1
    logical,intent(in)                  :: adj


    !===========================================
    ! kernels for fft and ifft
    !===========================================
    ! if adjoint, then do ifft
    if (adj) then
      ! loop over second axis (to ifft along first axis)
      do ii=1,n2
        ! perform ifftshift
        out1(1:(n1+1)/2) = cdata(n1-(n1+1)/2+1:n1,ii)
        out1((n1+1)/2+1:n1) = cdata(1:n1-(n1+1)/2,ii)
        ! perform ifft
        call sfftw_execute(fft_inv1)
        ! write out to cdata
        cdata(:,ii) = in1
      end do
    ! if forward, then do fft
    else
      do ii=1,n2
        ! read in cdata
        in1 = cdata(:,ii)
        ! perform fft
        call sfftw_execute(fft_for1)
        ! perform fftshift
        cdata(n1-(n1+1)/2+1:n1,ii) = out1(1:(n1+1)/2)
        cdata(1:n1-(n1+1)/2,ii) = out1((n1+1)/2+1:n1)
      end do
    end if

    ! scale cdata
    cdata = sign1*cdata/sqrt(real(n1))

  end subroutine




  !=============================================================
  ! fft on second axis
  !=============================================================
  subroutine fft_axis2(adj,sign2,cdata)

    !===========================================
    ! input local variables
    !===========================================
    integer                             :: ii
    complex,allocatable,dimension (:,:) :: cdata
    real                                :: sign2
    logical,intent(in)                  :: adj


    !===========================================
    ! kernels for fft and ifft
    !===========================================
    ! if adjoint, then do ifft
    if (adj) then
      ! loop over first axis (to ifft along first axis)
      do ii=1,n1
        ! perform ifftshift
        out2(1:(n2+1)/2) = cdata(ii,n2-(n2+1)/2+1:n2)
        out2((n2+1)/2+1:n2) = cdata(ii,1:n2-(n2+1)/2)
        ! perform ifft
        call sfftw_execute(fft_inv2)
        ! write out to cdata
        cdata(ii,:) = in2
      end do
    ! if forward, then do fft
    else
      do ii=1,n1
        ! read in cdata
        in2 = cdata(ii,:)
        ! perform fft
        call sfftw_execute(fft_for2)
        ! perform fftshift
        cdata(ii,n2-(n2+1)/2+1:n2) = out2(1:(n2+1)/2)
        cdata(ii,1:n2-(n2+1)/2) = out2((n2+1)/2+1:n2)
      end do
    end if

    ! scale cdata
    cdata = sign2*cdata/sqrt(real(n2))

  end subroutine


end module
