program jcChecker_perturb
  ! Program to create checkerboard
  ! Updated April 2015:
  !   - add percent perturbation to final velocity model
  !   - options include sinusoid or sharp edges
  !
  ! Inputs:
  !   1) history file containing (final) velocity model
  !   2) pturb: maximum percent perturbation
  !   3) opt: [1] sharp edges
  !            2  sinusoidal pattern
  !   4) lcx,lcx: length of checkers (in actual units)
  !
  ! Output:
  !   1) history file of perturbed checkers
  !
  ! Jason Chang
  ! February 2015

  !===================================
  ! standard declarations
  !===================================
  use sep
  implicit none


  !===================================
  ! declare variables
  !===================================
  ! program parameters
  integer                         :: ii,jj,kk
  integer                         :: indx1,indx2
  real,parameter                  :: pi=3.14159265359

  ! model parameters
  integer                         :: n1,n2,n3
  real                            :: d1,d2,d3,o1,o2,o3
  real,allocatable,dimension(:,:) :: model1,model2,pp

  ! checker parameters
  integer                         :: nxc,nyc,dxc,dyc
  integer                         :: nextrax,nextray
  integer                         :: opt
  real                            :: lcx,lcy,coord
  real                            :: pturb,val1,val2,temp
  real                            :: totx,toty,extrax,extray


  !===================================
  ! open I/O
  !===================================
  call sep_init()

  ! grab model domain parameters
  call from_history('n1',n1,1 )
  call from_history('o1',o1,0.)
  call from_history('d1',d1,1.)
  call from_history('n2',n2,1 )
  call from_history('o2',o2,0.)
  call from_history('d2',d2,1.)
  call from_history('n3',n3,1 )
  call from_history('o3',o3,0.)
  call from_history('d3',d3,1.)

  ! check parameters
  call from_param('lcx',lcx,1.)
  call from_param('lcy',lcy,1.)
  call from_param('opt',opt,1)
  call from_param('pturb',pturb,0.05)

  ! calculate values
  val1 = (1+pturb)
  val2 = (1-pturb)

  ! calculate checker sizes
  ! total length of model domain
  totx = n1*d1
  toty = n2*d2
  ! determine number of checkers in each direction (rounded down)
  nxc = totx/lcx
  nyc = toty/lcy
  dxc = lcx/d1
  dyc = lcy/d2
  ! work out extra samples at ends
  extrax = mod(totx,lcx)
  extray = mod(toty,lcy)
  nextrax = 1+extrax/d1
  nextray = 1+extray/d2
  

  !===================================
  ! initialize
  !===================================
  allocate(model1(n1,dyc))
  allocate(model2(n1,n2))
  allocate(pp(n1,n2))


  !===================================
  ! kernel
  !===================================
  ! loop over all models
  do kk=1,n3
    ! sharp-edged checkerboard
    if (opt.eq.1) then
      do ii=1,nyc
        ! read in model
        call sreed('in',model1,4*n1*dyc)
        do jj=1,nxc,2
          indx1 = 1+(jj-1)*dxc
          indx2 = jj*dxc
          model1(indx1:indx2,:) = model1(indx1:indx2,:)*val1
        end do
        do jj=2,nxc,2
          indx1 = 1+(jj-1)*dxc
          indx2 = jj*dxc
          model1(indx1:indx2,:) = model1(indx1:indx2,:)*val2
        end do
        ! extra bit
        if (extrax.ne.0.) then
          do jj=indx2+1,n1
            model1(jj,:) = model1(jj,:)*val1
          end do
        end if
      
        ! switch values
        temp = val1
        val1 = val2
        val2 = temp

        ! write out
        call srite('out',model1,4*n1*dyc)
      end do

      ! extra bits
      if (extray.ne.0.) then
        do jj=1,nxc,2
          indx1 = 1+(jj-1)*dxc
          indx2 = jj*dxc
          model1(indx1:indx2,:) = model1(indx1:indx2,:)*val1
        end do
        do jj=2,nxc,2
          indx1 = 1+(jj-1)*dxc
          indx2 = jj*dxc
          model1(indx1:indx2,:) = model1(indx1:indx2,:)*val2
        end do
        ! extra bit
        if (extrax.ne.0.) then
          do jj=indx2+1,n1
            model1(jj,:) = model1(jj,:)*val1
          end do
        end if

        ! write out
        call srite('out',model1(:,1:nextray),4*n1*nextray)
      end if
    end if


    ! sinusoidal option
    if (opt.eq.2) then
      ! read in model
      call sreed('in',model2,4*n1*n2)
      ! calculate sinusoid along axis1
      do ii=1,n1
        coord = (ii-0.5)*d1
        pp(ii,:) = sin(pi*coord/lcx)
      end do
      ! calculate sinusoid along axis2
      do ii=1,n2
        coord = (ii-0.5)*d2
        pp(:,ii) = pp(:,ii)*sin(pi*coord/lcy)
      end do
      pp = 1+pp*pturb
      ! write out
      call srite('out',pp*model2,4*n1*n2)
    end if
  end do


  !===================================
  ! write out
  !===================================
  call to_history('n1',n1)
  call to_history('n2',n2)
  call to_history('n3',n3)
  call to_history('o1',o1)
  call to_history('o2',o2)
  call to_history('o3',o3)
  call to_history('d1',d1)
  call to_history('d2',d2)
  call to_history('d3',d3)


  !===================================
  ! clean up
  !===================================
  call sep_close()
  deallocate(model1,model2,pp)
  call exit(0)

end program
