program jcFK
  ! Program to compute FK spectrum of data
  ! Calls upon fourier module
  !
  ! Input:
  !   1) history file (n1: time, n2:distance, n3: number of gathers)
  !     - this is a complex input
  !	- assumes that both axes begin with the origin/zero component
  !
  ! Output:
  !   1) history file (n1: temporal frequency, n2: spatial frequency, n3: number of gathers)
  !     - this is a complex output
  !
  ! Jason Chang
  ! February 2014

  !===========================================
  ! standard declarations
  !===========================================
  use sep
  use fourier_mod

  implicit none
 

  !===========================================
  ! declare variables
  !===========================================
  integer                               :: ii
  integer                               :: n1,n2,n3
  integer                               :: nw,nk
  real                                  :: d1,d2,d3,o1,o2,o3
  real                                  :: dw,dk,ow,ok
  complex,allocatable,dimension (:,:)   :: cdata


  !===========================================
  ! open I/O
  !===========================================
  call sep_init()

  ! input history file parameters
  call from_history('n1',n1,1 ) !time
  call from_history('d1',d1,1.)
  call from_history('o1',o1,0.)
  call from_history('n2',n2,1 ) !space
  call from_history('d2',d2,1.)
  call from_history('o2',o2,0.)
  call from_history('n3',n3,1 ) !number of gathers
  call from_history('d3',d3,1.)
  call from_history('o3',o3,0.)

  ! calculate fourier units
  nw = n1
  dw = 1/((n1-1)*d1)
  ow = -1./(2*d1)
  nk = n2
  dk = 1/((n2-1)*d2)
  ok = -1./(2*d2)

  ! output history file parameters
  call to_history('n1',nw) !temporal frequency
  call to_history('d1',dw)
  call to_history('o1',ow)
  call to_history('n2',nk) !spatial frequency
  call to_history('d2',dk)
  call to_history('o2',ok)
  call to_history('n3',n3) !number of gathers
  call to_history('d3',d3)
  call to_history('o3',o3)
  call to_history('esize',8) ! output is complex
  call to_history('label1','frequency [1/s]')
  call to_history('label2','wavenumber [1/km]')

  call sep_close()


  !===========================================
  ! allocate
  !===========================================
  allocate(cdata(n1,n2)); cdata=0.


  !===========================================
  ! initialize module
  !===========================================
  call fft_init(n1,n2)


  !===========================================
  ! kernel
  !===========================================
  do ii = 1,n3

    ! read in data
    call sreed('in',cdata,8*n1*n2)

    ! perform fft on time axis
    call fft_axis1(.false.,1.,cdata)

    ! perform fft on space axis
    call fft_axis2(.false.,-1.,cdata)

    ! write out data
    call srite('out',cdata,8*n1*n2)

  end do


  !===========================================
  ! clean up
  !===========================================
  call fft_close()
  deallocate(cdata)

end program
