program jcLCurve
  ! Program for plotting L-curves
  ! Assumes Laplacian model regularization
  ! Output not in log scale
  !
  ! Inputs:
  !   1) model: history file containing model output
  !   2) error: history file containing final data residual information
  !
  ! Outputs:
  !   1) history file
  !      - n1: data residual
  !      - n2: model smoothness values
  !
  ! Jason Chang
  ! March 2015


  !===================================
  ! standard declarations
  !===================================
  use sep
  use laplacian2dOMP_mod
  implicit none


  !===================================
  ! declare variables
  !===================================
  ! program parameters
  integer                         :: ii,stat

  ! input model parameters
  integer                         :: ngx,ngy,neps
  real                            :: oeps,deps
  real                            :: errm
  real,allocatable,dimension(:)   :: modin,modout

  ! error parameters
  integer                         :: ne1,ne2,ne3
  real                            :: oe1,oe2,oe3,de1,de2,de3
  real                            :: errd
  real,allocatable,dimension(:,:) :: error


  !===================================
  ! open I/O
  !===================================
  call sep_init()

  ! input history file parameters
  if (exist_file('error')) then
    call from_aux('error','n1',ne1,1 )
    call from_aux('error','o1',oe1,0.)
    call from_aux('error','d1',de1,1.)
    call from_aux('error','n2',ne2,1 )
    call from_aux('error','o2',oe2,0.)
    call from_aux('error','d2',de2,1.)
    call from_aux('error','n3',ne3,1 )
    call from_aux('error','o3',oe3,0.)
    call from_aux('error','d3',de3,1.)
  end if

  ! model parameters
  if (exist_file('model')) then
    call from_aux('model','n1',ngx ,1 )
    call from_aux('model','n2',ngy ,1 )
    call from_aux('model','n3',neps,1 )
    call from_aux('model','o3',oeps,0.)
    call from_aux('model','d3',deps,1.)
  end if

  ! check that axes match (number of regularization parameters)
  if (ne3.ne.neps) call seperr('n of regularization parameters do not match')
  

  !===================================
  ! initialization
  !===================================
  allocate(error(ne1,ne2))
  allocate(modin(ngx*ngy))
  allocate(modout(ngx*ngy))
  call laplacian2dOMP_init(ngx,ngy)


  !===================================
  ! kernel
  !===================================
  do ii=1,neps
    ! read in error information
    call sreed('error',error,4*ne1*ne2)
    ! calculate L2-norm of error
    errd = sqrt(error(ne1,1))

    ! read in model imformation
    call sreed('model',modin,4*ngx*ngy)
    ! calculate laplacian of model
    !call laplacian2dOMP_for(modin,modout)
    stat = laplacian2dOMP_lop(.false.,.false.,modin,modout)
    ! calculate L2-norm of laplacian of model
    errm = sqrt(dot_product(modout,modout))

    ! write out
    call srite('out',errm,4)
    call srite('out',errd,4)

  end do
  


  !===================================
  ! write out
  !===================================
  ! output history file parameters
  call to_history('n1',2 )
  call to_history('o1',0.)
  call to_history('d1',1.)
  call to_history('n2',neps)
  call to_history('o2',oeps)
  call to_history('d2',deps)


  !===================================
  ! clean up
  !===================================
  deallocate(error,modin,modout)
  call laplacian2dOMP_close()
  call sep_close()
  call exit(0)

end program
