program jcRMSerror
  ! Program to calculate RMS error in travel times
  ! Calculates percent error based on:
  ! |tsyn-tobs|/tobs
  !
  ! Inputs:
  !   1) history file containing final model
  !   2) tfile: file containing observed traveltimes
  !       - assumes thdde long beach format for keys
  !
  ! Outputs:
  !   1) NOT SURE YET
  !
  ! Jason Chang
  ! April 2015


  !===================================
  ! standard declarations
  !===================================
  use sep
  use tomo_straightray_mod
  implicit none


  !===================================
  ! declare variables
  !===================================
  ! program parameters
  integer                         :: ii

  ! model parameters
  integer                         :: n1,n2
  real                            :: o1,o2,d1,d2
  real,allocatable,dimension(:)   :: model

  ! header parameters
  integer                         :: nh1,nh2
  integer                         :: rxkey,rykey,sxkey,sykey
  integer                         :: rlkey,rpkey,slkey,spkey
  integer                         :: ttkey,snrkey
  real,allocatable,dimension(:,:) :: header

  ! threshold parameters
  integer                         :: nkeep,ntraces
  integer                         :: rpoint,rline,spoint,sline
  real                            :: snr,minr,maxr,dist

  ! output parameters
  real                            :: rms
  real,allocatable,dimension(:)   :: dat


  !===================================
  ! open I/O
  !===================================
  call sep_init()

  ! model parameters
  call from_history('n1',n1,1 )
  call from_history('n2',n2,1 )
  call from_history('o1',o1,0.)
  call from_history('o2',o2,0.)
  call from_history('d1',d1,1.)
  call from_history('d2',d2,1.)

  ! header file parameters
  call from_param('rxkey' ,rxkey ,1 )
  call from_param('rykey' ,rykey ,2 )
  call from_param('sxkey' ,sxkey ,3 )
  call from_param('sykey' ,sykey ,4 )
  call from_param('rlkey' ,rlkey ,5 )
  call from_param('rpkey' ,rpkey ,6 )
  call from_param('slkey' ,slkey ,7 )
  call from_param('spkey' ,spkey ,8 )
  call from_param('ttkey' ,ttkey ,9 )
  call from_param('snrkey',snrkey,10)

  ! threshold parameters
  call from_param('snr' ,snr ,-1.)
  call from_param('minr',minr,-1.)
  call from_param('maxr',maxr,-1.)

  if (exist_file('hff')) then
    call from_aux('hff','n1',nh1,1 )
    call from_aux('hff','n2',nh2,1 )
  else
    call seperr('Need header history file')
  end if


  !===================================
  ! initialize
  !===================================
  allocate(model(n1*n2))
  allocate(header(nh1,nh2))


  !===================================
  ! QC/clean out traces
  !===================================
  ! read in header information
  call sreed('hff',header,4*nh1*nh2)
  ! get rid of reciprocal traces
  ! needs to be based on line and point coordinates
  ntraces = nh2
  nkeep = 0
  do ii = 1,ntraces
    ! grab all (line,point) coordinates
    rline  = header(rlkey,ii)
    rpoint = header(rpkey,ii)
    sline  = header(slkey,ii)
    spoint = header(spkey,ii)
    ! take only the lower values
    if (rline.gt.sline) cycle
    if (rline.eq.sline .and. rpoint.gt.spoint) cycle
    nkeep = nkeep+1
    ! move trace up in vector
    header(:,nkeep) = header(:,ii)
  end do
  ! zero out the rest of the vector
  ntraces = nkeep
  if (ntraces.ne.nh2) header(:,ntraces+1:nh2) = 0.
write(0,*) "after non-reciprocal traces:",ntraces

  ! based on snr
  nkeep = 0
  do ii = 1,ntraces
    ! some sort of snr filter
    if (snr.ne.-1 .and. header(snrkey,ii).lt.snr) cycle
    nkeep = nkeep+1
    header(:,nkeep) = header(:,ii)
  end do
  ! zero out the rest of the vector
  ntraces = nkeep
  if (ntraces.ne.nh2) header(:,ntraces+1:nh2) = 0.
write(0,*) "after snr filter:",ntraces

  ! based on min and max distance between receivers
  nkeep = 0
  do ii = 1,ntraces
    ! some sort of distance filter
    dist = sqrt( (header(rxkey,ii)-header(sxkey,ii))**2. + (header(rykey,ii)-header(sykey,ii))**2. )
    if (minr.ne.-1 .and. minr.gt.dist) cycle
    if (maxr.ne.-1 .and. maxr.lt.dist) cycle
    nkeep = nkeep+1
    header(:,nkeep) = header(:,ii)
  end do
  ! zero out the rest of the vector
  ntraces = nkeep
  if (ntraces.ne.nh2) header(:,ntraces+1:nh2) = 0.
write(0,*) "after min/max distance:",ntraces



  !===================================
  ! kernel
  !===================================
  allocate(dat(ntraces))
  ! read in header and model
  call sreed('in',model,4*n1*n2)
  model = 1/model

  ! initialize the tomography
  call tomo_straightray_init(ntraces,header(rxkey,:),header(rykey,:),header(sxkey,:),header(sykey,:),n1,o1,d1,n2,o2,d2)

  ! create travel times
  call tomo_straightray_for(model,dat)

  ! calculate RMS error
  rms = 0
  do ii=1,ntraces
    rms = rms+(abs(dat(ii)-header(ttkey,ii))/header(ttkey,ii))**2
    !rms = rms+(abs(dat(ii)-header(ttkey,ii))/header(ttkey,ii))
  end do
  rms = sqrt(rms/ntraces)
  !rms = rms/ntraces
write(0,*) rms

  ! write out data
  call srite('out',dat,4*ntraces)


  !===================================
  ! write out
  !===================================
  call to_history('n1',ntraces)
  call to_history('n2',1  )
  call to_history('o2',0. )
  call to_history('d2',1. )


  !===================================
  ! clean up
  !===================================
  deallocate(model,header,dat)
  call sep_close()
  call tomo_straightray_close()
  call exit(0)

end program
