program jcRayLength
  ! Program for calculating cumulative ray lengths through grid cells
  ! For tomography purposes
  !
  ! Inputs:
  !   1) history file containing (rx,ry,sx,sy,time,snr)
  !   2) snr threshold
  !   3) minimum and maximum s-r distance threshold
  !   4) model domain parameters (ngx,dgx,ogx,ngy,dgy,ogy)
  !
  ! Outputs:
  !   1) history file with ray lengths in each grid cell
  !
  ! Jason Chang
  ! March 2015


  !===================================
  ! standard declarations
  !===================================
  use sep
  implicit none


  !===================================
  ! declare variables
  !===================================
  ! program parameters
  integer                         :: ii,jj
  integer                         :: irx,iry,isx,isy
  integer                         :: prev_ix,prev_iy,next_ix,next_iy
  integer                         :: xstep,ystep
  integer                         :: nkeep,ntraces
  real                            :: xoffset,yoffset
  real                            :: xmove1,ymove1,xmove2,ymove2
  real                            :: prev_xcoord,prev_ycoord,next_xcoord,next_ycoord
  real                            :: dist
  real                            :: m0
  real,allocatable,dimension(:)   :: distance,rx,ry,sx,sy

  ! input parameters
  integer                         :: n1,n2
  integer                         :: rxkey,rykey,sxkey,sykey
  integer                         :: rlkey,rpkey,slkey,spkey
  integer                         :: rline,rpoint,sline,spoint
  integer                         :: ttkey,snrkey
  real,allocatable,dimension(:,:) :: datain
  real,allocatable,dimension(:)   :: d

  ! output parameters
  integer                         :: ngx,ngy
  real                            :: ogx,ogy,dgx,dgy
  real,allocatable,dimension(:)   :: m

  ! threshold parameters
  real                            :: snr,minr,maxr


  !===================================
  ! open I/O
  !===================================
  call sep_init()

  ! input history file parameter
  call from_history('n1',n1,1)
  call from_history('n2',n2,1)

  ! input history file header keys
  call from_param('rxkey' ,rxkey ,1 )
  call from_param('rykey' ,rykey ,2 )
  call from_param('sxkey' ,sxkey ,3 )
  call from_param('sykey' ,sykey ,4 )
  call from_param('rxkey' ,rlkey ,5 )
  call from_param('rykey' ,rpkey ,6 )
  call from_param('sxkey' ,slkey ,7 )
  call from_param('sykey' ,spkey ,8 )
  call from_param('ttkey' ,ttkey ,9 )
  call from_param('snrkey',snrkey,10)

  ! threshold parameters
  call from_param('snr' ,snr ,-1.)
  call from_param('minr',minr,-1.)
  call from_param('maxr',maxr,-1.)

  ! output parameters
  call from_param('ngx',ngx,1 )
  call from_param('ogx',ogx,0.)
  call from_param('dgx',dgx,1.)
  call from_param('ngy',ngy,1 )
  call from_param('ogy',ogy,0.)
  call from_param('dgy',dgy,1.)

 
  !===================================
  ! read in information
  !===================================
  allocate(datain(n1,n2))
  call sreed('in',datain,4*n1*n2)


  !===================================
  ! QC / clean out traces
  !===================================
  ! get rid of reciprocal traces
  ! needs to be based on line and point coordinates
  ntraces = n2
  nkeep = 0
  do ii = 1,ntraces
    ! grab all (line,point) coordinates
    rline  = datain(rlkey,ii)
    rpoint = datain(rpkey,ii)
    sline  = datain(slkey,ii)
    spoint = datain(spkey,ii)
    ! take only the lower values
    if (rline.gt.sline) cycle
    if (rline.eq.sline .and. rpoint.gt.spoint) cycle
    nkeep = nkeep+1
    ! move trace up in vector
    datain(:,nkeep) = datain(:,ii)
  end do
  ! zero out the rest of the vector
  ntraces = nkeep
  if (ntraces.ne.n2) datain(:,ntraces+1:n2) = 0.
  write(0,*) "after non-reciprocal traces:",ntraces

  ! based on snr
  nkeep = 0
  do ii = 1,ntraces
    ! some sort of snr filter
    if (snr.ne.-1 .and. datain(snrkey,ii).lt.snr) cycle
    nkeep = nkeep+1
    datain(:,nkeep) = datain(:,ii)
  end do
  ! zero out the rest of the vector
  ntraces = nkeep
  if (ntraces.ne.n2) datain(:,ntraces+1:n2) = 0.
  write(0,*) "after snr filter:",ntraces

  ! based on min and max distance between receivers
  allocate(distance(ntraces))
  nkeep = 0
  do ii = 1,ntraces
    ! some sort of distance filter
    dist = sqrt( (datain(rxkey,ii)-datain(sxkey,ii))**2. + (datain(rykey,ii)-datain(sykey,ii))**2. )
    if (minr.ne.-1 .and. minr.gt.dist) cycle
    if (maxr.ne.-1 .and. maxr.lt.dist) cycle
    nkeep = nkeep+1
    datain(:,nkeep) = datain(:,ii)
    distance(nkeep) = dist
  end do
  ! zero out the rest of the vector
  ntraces = nkeep
  if (ntraces.ne.n2) datain(:,ntraces+1:n2) = 0.
  write(0,*) "after min/max distance:",ntraces


  !===================================
  ! initialize
  !===================================
  ! allocate and assign values
  allocate(m(ngx*ngy)) ; m = 0.
  allocate(d(ntraces)) ; d = datain(ttkey,1:ntraces)
  allocate(rx(ntraces)); rx = datain(rxkey,1:ntraces)
  allocate(ry(ntraces)); ry = datain(rykey,1:ntraces)
  allocate(sx(ntraces)); sx = datain(sxkey,1:ntraces)
  allocate(sy(ntraces)); sy = datain(sykey,1:ntraces)

  ! calculate initial homogenous model
  m0 = 0.
  do ii = 1,ntraces
    m0 = m0+(d(ii)/distance(ii))
  end do
  m0 = m0/ntraces
  write(0,*) "average slowness:",m0


  !===================================
  ! kernel
  !===================================
  do ii = 1,ntraces
    ! determine where the source and receiver are in model domain indices
    irx = 1+int((rx(ii)-ogx)/dgx)
    iry = 1+int((ry(ii)-ogy)/dgy)
    isx = 1+int((sx(ii)-ogx)/dgx)
    isy = 1+int((sy(ii)-ogy)/dgy)

    ! determine which direction to step in (from source to receiver)
    ! indicate which general directions to move in
    if (sx(ii).lt.rx(ii)) then
      xstep = 1
    elseif (sx(ii).gt.rx(ii)) then
      xstep = -1
    else
      xstep = 0
    end if
    if (sy(ii).lt.ry(ii)) then
      ystep = 1
    elseif (sy(ii).gt.ry(ii)) then
      ystep = -1
    else
      ystep = 0
    end if
    ! take care of situation where rays are along boundaries
    if (rx(ii).eq.(ogx+(irx-1)*dgx) .and. xstep.eq. 1) irx = irx-1
    if (ry(ii).eq.(ogy+(iry-1)*dgy) .and. ystep.eq. 1) iry = iry-1
    if (sx(ii).eq.(ogx+(isx-1)*dgx) .and. xstep.eq.-1) isx = isx-1
    if (sy(ii).eq.(ogy+(isy-1)*dgy) .and. ystep.eq.-1) isy = isy-1
    ! break out of loop if indices are out of range
    if (irx.lt.1 .or. irx.gt.ngx) cycle
    if (isx.lt.1 .or. isx.gt.ngx) cycle
    if (iry.lt.1 .or. iry.gt.ngy) cycle
    if (isy.lt.1 .or. isy.gt.ngy) cycle
    ! set initial x and y indices and coordinates
    prev_ix = isx
    prev_iy = isy
    prev_xcoord = sx(ii)
    prev_ycoord = sy(ii)

    ! loop until receiver indices are reached
    do while (prev_ix.ne.irx .or. prev_iy.ne.iry)
      ! calculate total x and y offsets (will effectively provide a slope)
      xoffset = rx(ii)-prev_xcoord
      yoffset = ry(ii)-prev_ycoord
      ! input lengths of potential movements
      ! if we move along x, then compute corresponding y
      if (xstep.eq.1) then
        xmove1 = (ogx+(prev_ix*dgx))-prev_xcoord
        ymove1 = xmove1*(yoffset/(xoffset))
      else if (xstep.eq.-1) then
        xmove1 = (ogx+(prev_ix-1)*dgx)-prev_xcoord
        ymove1 = xmove1*(yoffset/(xoffset))
      else
        xmove1 = 0
        ymove1 = 1/0.
      end if
      ! if we move along y, then compute corresponding x
      if (ystep.eq.1) then
        ymove2 = (ogy+(prev_iy*dgy))-prev_ycoord
        xmove2 = ymove2*(xoffset/(yoffset))
      else if (ystep.eq.-1) then
        ymove2 = (ogy+(prev_iy-1)*dgy)-prev_ycoord
        xmove2 = ymove2*(xoffset/(yoffset))
      else
        ymove2 = 0
        xmove2 = 1/0.
      end if

      ! determine which direction to walk in
      ! step in the +/- x direction
      if (abs(xmove2).gt.abs(xmove1) .and. abs(ymove2).gt.abs(ymove1)) then
        ! update both step index and actual coordinate
        next_ix = prev_ix+xstep
        next_iy = prev_iy
        if (xstep.eq. 1) next_xcoord = ogx+(next_ix-1)*dgx
        if (xstep.eq.-1) next_xcoord = ogx+(next_ix  )*dgx
        next_ycoord = prev_ycoord+(next_xcoord-prev_xcoord)*(yoffset/(xoffset))
      ! step in the +/- y direction
      else if (abs(xmove1).gt.abs(xmove2) .and. abs(ymove1).gt.abs(ymove2)) then
        ! update both step index and actual coordinate
        next_iy = prev_iy+ystep
        next_ix = prev_ix
        if (ystep.eq. 1) next_ycoord = ogy+(next_iy-1)*dgy
        if (ystep.eq.-1) next_ycoord = ogy+(next_iy  )*dgy
        next_xcoord = prev_xcoord+(next_ycoord-prev_ycoord)*(xoffset/(yoffset))
      ! step in the +/- x and y directions
      else
        ! update both step index and actual coordinate
        ! scenario where moving perfectly horizontally or vertically
        next_ix = prev_ix
        next_iy = prev_iy
        next_xcoord = prev_xcoord
        next_ycoord = prev_ycoord
        ! scenario where moving diagonally through model domain nodes
        if (prev_ix.ne.irx) then
          next_ix = prev_ix+xstep
          if (xstep.eq. 1) next_xcoord = ogx+((next_ix-1)*dgx)
          if (xstep.eq.-1) next_xcoord = ogx+(next_ix*dgx)
        end if
        if (prev_iy.ne.iry) then
          next_iy = prev_iy+ystep
          if (ystep.eq. 1) next_ycoord = ogy+((next_iy-1)*dgy)
          if (ystep.eq.-1) next_ycoord = ogy+(next_iy*dgy)
        end if
      end if
      ! calculate length in cell and place in appropriate model index
      dist = sqrt((next_xcoord-prev_xcoord)**2 + (next_ycoord-prev_ycoord)**2)
      jj = ((prev_iy-1)*ngx)+prev_ix
      ! integrate / matrix multiply
      m(jj) = m(jj)+dist

      ! cycle through coordinates and indices
      prev_xcoord = next_xcoord
      prev_ycoord = next_ycoord
      prev_ix = next_ix
      prev_iy = next_iy

    end do

    ! once receiver cell reached, finish the last bit within that cell
    !dist = sqrt((next_xcoord-prev_xcoord)**2 + (next_ycoord-prev_ycoord)**2)
    dist = sqrt((rx(ii)-prev_xcoord)**2 + (ry(ii)-prev_ycoord)**2)
    jj = ((prev_iy-1)*ngx)+prev_ix
    ! integrate / matrix multiply
    m(jj) = m(jj)+dist

  end do

  ! write out
  call srite('out',m,4*ngx*ngy)

  !===================================
  ! write out
  !===================================
  call to_history('n1',ngx)
  call to_history('o1',ogx)
  call to_history('d1',dgx)
  call to_history('n2',ngy)
  call to_history('o2',ogy)
  call to_history('d2',dgy)


  !===================================
  ! clean up
  !===================================
  deallocate(datain,m,d,distance,rx,ry,sx,sy)
  call sep_close()
  call exit(0)

  
end program
