program jcStraightRayTomoOMP
  ! Program for performing straight ray tomography
  ! Modification: i
  !   - works with slowness perturbations (February 2015)
  !   - activated reciprocal traces (February 2015)
  !
  ! Inputs:
  !   1) history file containing (rx,ry,sx,sy,time,snr)
  !   2) snr threshold
  !   3) minimum and maximum s-r distance threshold
  !   4) regularization parameter
  !   5) model domain parameters (ngx,dgx,ogx,ngy,dgy,ogy)
  !   6) residual file
  !   7) startm: starting model
  !
  ! Outputs:
  !   1) history file containing slowness values at each grid point
  !   2) residual plot
  !
  ! Jason Chang
  ! November 2014

  !===================================
  ! standard declarations
  !===================================
  use sep
  use random_mod
  use tomo_straightrayOMP_mod
  use laplacian2dOMP_mod
  use regsolver_mod
  use cgstep_mod
  implicit none

  
  !===================================
  ! declare variables
  !===================================
  ! program parameters
  integer                         :: ii
  integer                         :: nkeep,ntraces
  integer                         :: niter
  real                            :: dist,m0,eps,mavg
  real,allocatable,dimension(:)   :: m00,distance
  logical                         :: inv
  
  ! dot_test parameters
  integer                         :: stat
  real                            :: rand
  real                            :: dot1,dot2
  real,allocatable,dimension(:)   :: m1,m2,d1,d2
  logical                         :: dot_test

  ! input parameters
  integer                         :: n1,n2
  integer                         :: rxkey,rykey,sxkey,sykey
  integer                         :: rlkey,rpkey,slkey,spkey
  integer                         :: rline,rpoint,sline,spoint
  integer                         :: ttkey,snrkey
  real,allocatable,dimension(:,:) :: datain
  real,allocatable,dimension(:)   :: d

  ! output parameters
  integer                         :: ngx,ngy
  real                            :: ogx,ogy,dgx,dgy
  real,allocatable,dimension(:)   :: m,res,resd
  real,allocatable,dimension(:,:) :: error

  ! threshold parameters
  real                            :: snr,minr,maxr

  
  !===================================
  ! open I/O
  !===================================
  call sep_init()

  ! dot test parameters
  call from_param('dot_test',dot_test,.false.)
  call from_param('inv'     ,inv     ,.true.)

  ! input history file parameters
  call from_history('n1',n1,1)
  call from_history('n2',n2,1)

  ! input history file header keys
  call from_param('rxkey' ,rxkey ,1 )
  call from_param('rykey' ,rykey ,2 )
  call from_param('sxkey' ,sxkey ,3 )
  call from_param('sykey' ,sykey ,4 )
  call from_param('rxkey' ,rlkey ,5 )
  call from_param('rykey' ,rpkey ,6 )
  call from_param('sxkey' ,slkey ,7 )
  call from_param('sykey' ,spkey ,8 )
  call from_param('ttkey' ,ttkey ,9 )
  call from_param('snrkey',snrkey,10)

  ! threshold parameters
  call from_param('snr' ,snr ,-1.)
  call from_param('minr',minr,-1.)
  call from_param('maxr',maxr,-1.)

  ! optimization operators
  call from_param('niter',niter,1)
  call from_param('eps',eps,0.4)

  ! read output parameters
  call from_param('ngx',ngx,1 )
  call from_param('ogx',ogx,0.)
  call from_param('dgx',dgx,1.)
  call from_param('ngy',ngy,1 )
  call from_param('ogy',ogy,0.)
  call from_param('dgy',dgy,1.)

  ! determine shape of output
  call to_history('n1',ngx)
  call to_history('o1',ogx)
  call to_history('d1',dgx)
  call to_history('n2',ngy)
  call to_history('o2',ogy)
  call to_history('d2',dgy)

  ! optional residual
  if (exist_file('error')) then
    call to_history('n1',niter,'error')
    call to_history('n2',4    ,'error')
  end if


  
  !===================================
  ! read in information
  !===================================
  allocate(datain(n1,n2))
  call sreed('in',datain,4*n1*n2)


  !===================================
  ! QC / clean out traces
  !===================================
  ! get rid of reciprocal traces
  ! needs to be based on line and point coordinates
  ntraces = n2
  nkeep = 0
  do ii = 1,ntraces
    ! grab all (line,point) coordinates
    rline  = datain(rlkey,ii)
    rpoint = datain(rpkey,ii)
    sline  = datain(slkey,ii)
    spoint = datain(spkey,ii)
    ! take only the lower values
    if (rline.gt.sline) cycle
    if (rline.eq.sline .and. rpoint.gt.spoint) cycle
    nkeep = nkeep+1
    ! move trace up in vector
    datain(:,nkeep) = datain(:,ii)
  end do
  ! zero out the rest of the vector
  ntraces = nkeep
  if (ntraces.ne.n2) datain(:,ntraces+1:n2) = 0.
write(0,*) "after non-reciprocal traces:",ntraces

  ! based on snr
  nkeep = 0
  do ii = 1,ntraces
    ! some sort of snr filter
    if (snr.ne.-1 .and. datain(snrkey,ii).lt.snr) cycle
    nkeep = nkeep+1
    datain(:,nkeep) = datain(:,ii)
  end do
  ! zero out the rest of the vector
  ntraces = nkeep
  if (ntraces.ne.n2) datain(:,ntraces+1:n2) = 0.
write(0,*) "after snr filter:",ntraces

  ! based on min and max distance between receivers
  allocate(distance(ntraces))
  nkeep = 0
  do ii = 1,ntraces
    ! some sort of distance filter
    dist = sqrt( (datain(rxkey,ii)-datain(sxkey,ii))**2. + (datain(rykey,ii)-datain(sykey,ii))**2. )
    if (minr.ne.-1 .and. minr.gt.dist) cycle
    if (maxr.ne.-1 .and. maxr.lt.dist) cycle
    nkeep = nkeep+1
    datain(:,nkeep) = datain(:,ii)
    distance(nkeep) = dist
  end do
  ! zero out the rest of the vector
  ntraces = nkeep
  if (ntraces.ne.n2) datain(:,ntraces+1:n2) = 0.
write(0,*) "after min/max distance:",ntraces

  !===================================
  ! get ready for operators
  !===================================
  ! allocate and assign values
  allocate(m(ngx*ngy)); m = 0.
  allocate(d(ntraces)); d = datain(ttkey,1:ntraces)
  allocate(error(niter,4))
  allocate(res(ntraces))
  allocate(resd(ntraces))

  ! calculate initial homogeneous model
  m0 = 0.
  do ii = 1,ntraces
    m0 = m0+(d(ii)/distance(ii))
  end do
  mavg = m0/ntraces
  m = mavg
  allocate(m00(ngx*ngy))
  if (exist_file('startm')) then
    call sreed('startm',m00,4*ngx*ngy)
    m00 = 1/m00
  else
    m00=m
    write(0,*) "starting slowness:",mavg
  end if

  ! modify data
  do ii = 1,ntraces
    d(ii) = d(ii)-(mavg*distance(ii))
  end do

  ! initalize
  call tomo_straightrayOMP_init(ntraces,datain(rxkey,1:ntraces),datain(rykey,1:ntraces),datain(sxkey,1:ntraces),datain(sykey,1:ntraces),ngx,ogx,dgx,ngy,ogy,dgy)
  call laplacian2dOMP_init(ngx,ngy)
  
  !===================================
  ! dot test option
  !===================================
  if (dot_test) then
    ! test tomography operator
    call random_init(1)

    ! allocate
    allocate(d1(ntraces))
    allocate(d2(ntraces))
    allocate(m1(ngx*ngy))
    allocate(m2(ngx*ngy))

    ! fill vectors with random values
    do ii = 1,ntraces
      call random(rand)
      d2(ii) = rand
    end do
    do ii = 1,ngx*ngy
      call random(rand)
      m1(ii) = rand
    end do

    ! run dot product test
    dot1 = 0.
    dot2 = 0.
    stat = tomo_straightrayOMP_lop(.false.,.false.,m1,d1)
    stat = tomo_straightrayOMP_lop(.true. ,.false.,m2,d2)
    do ii = 1,ntraces
      dot1 = dot1+(d1(ii)*d2(ii))
    end do
    do ii = 1,ngx*ngy
      dot2 = dot2+(m1(ii)*m2(ii))
    end do

    ! write out results
    write(0,*) "dot1:",dot1
    write(0,*) "dot2:",dot2

    ! clean up
    deallocate(d1,d2,m1,m2)

 
    ! test laplacian regularization operator
    allocate(d1(ngx*ngy))
    allocate(d2(ngx*ngy))
    allocate(m1(ngx*ngy))
    allocate(m2(ngx*ngy))

    ! fill vectors with random values
    do ii = 1,ngx*ngy
      call random(rand)
      d2(ii) = rand      
    end do
    do ii = 1,ngx*ngy
      call random(rand)
      m1(ii) = rand      
    end do

    ! run dot product test
    dot1 = 0.
    dot2 = 0.
    stat = laplacian2dOMP_lop(.false.,.false.,m1,d1)
    stat = laplacian2dOMP_lop(.true. ,.false.,m2,d2)
    do ii = 1,ngx*ngy
      dot1 = dot1+(d1(ii)*d2(ii))
    end do
    do ii = 1,ngx*ngy
      dot2 = dot2+(m1(ii)*m2(ii))
    end do

    ! write out results
    write(0,*) "dot1:",dot1
    write(0,*) "dot2:",dot2

    ! clean up
    deallocate(d1,d2,m1,m2)
  end if


  !===================================
  ! kernel
  !===================================
  if (inv) then
    call regsolver(m=m,d=d,Fop=tomo_straightrayOMP_lop,Aop=laplacian2dOMP_lop,stepper=cgstep,niter=niter,nAop=ngx*ngy,eps=eps,m0=m00,resd=resd,err=error,verb=.true.)
  else
    stat = tomo_straightrayOMP_lop(.true.,.false.,m,d)
  end if
  call srite('out',1/(m+mavg),4*ngx*ngy)
  if (exist_file('error')) call srite('error',error,4*niter*4)
  if (exist_file('resd'))  call srite('resd',resd,4*ntraces)

  ! optional data residual
  if (exist_file('resd')) then
    call to_history('n1',ntraces,'resd')
    call to_history('n2',1 ,'resd')
  end if

  !===================================
  ! clean up
  !===================================
  deallocate(datain,m,d,m00,res,distance,resd,error)
  call sep_close()
  call tomo_straightrayOMP_close()
  call laplacian2dOMP_close()
  call cgstep_close()
  call exit(0)

end program
