program jcSyntheticTT_noise
  ! Program to calculate synthetic traveltimes
  ! Applies RMS percent error as standard deviation method
  !   - adds noise to travel measured times based on gaussian distribution
  !
  ! Inputs:
  !   1) history file containing final model (true or checkerboard)
  !   2) hff: file containing observed traveltimes/picks
  !       - assumes the long beach format for keys
  !   3) opt: [1] add no noise
  !            2  add noise based on RMS percent error
  !   4) seed: for random
  !   5) rms: percent error (obtained from jcRMSerror)
  !
  ! Outputs:
  !   1) history file of travel times
  !
  ! Jason Chang
  ! April 2015


  !===================================
  ! standard declarations
  !===================================
  use sep
  use random_mod
  use tomo_straightray_mod
  implicit none


  !===================================
  ! declare variables
  !===================================
  ! program parameters
  integer                         :: ii,jj
  real,parameter                  :: pi=3.14159265359

  ! model parameters
  integer                         :: n1,n2,n3
  real                            :: o1,o2,o3,d1,d2,d3
  real,allocatable,dimension(:)   :: model

  ! header parameters
  integer                         :: nh1,nh2
  integer                         :: rxkey,rykey,sxkey,sykey
  integer                         :: rlkey,rpkey,slkey,spkey
  integer                         :: ttkey,snrkey
  real,allocatable,dimension(:,:) :: header

  ! threshold parameters
  integer                         :: nkeep,ntraces
  integer                         :: rpoint,rline,spoint,sline
  real                            :: snr,minr,maxr,dist

  ! output parameters
  real,allocatable,dimension(:)   :: dat
  real,allocatable,dimension(:,:)  :: datout

  ! noise parameters
  integer                         :: opt,seed
  real                            :: rms,stdev1,stdev2,noise1,noise2,rand1,rand2
  

  !===================================
  ! open I/O
  !===================================
  call sep_init()

  ! model parameters
  call from_history('n1',n1,1 )
  call from_history('n2',n2,1 )
  call from_history('n3',n3,1 )
  call from_history('o1',o1,0.)
  call from_history('o2',o2,0.)
  call from_history('o3',o3,0.)
  call from_history('d1',d1,1.)
  call from_history('d2',d2,1.)
  call from_history('d3',d3,1.)

  ! header file parameters
  call from_param('rxkey' ,rxkey ,1 )
  call from_param('rykey' ,rykey ,2 )
  call from_param('sxkey' ,sxkey ,3 )
  call from_param('sykey' ,sykey ,4 )
  call from_param('rlkey' ,rlkey ,5 )
  call from_param('rpkey' ,rpkey ,6 )
  call from_param('slkey' ,slkey ,7 )
  call from_param('spkey' ,spkey ,8 )
  call from_param('ttkey' ,ttkey ,9 )
  call from_param('snrkey',snrkey,10)

  ! threshold parameters
  call from_param('snr' ,snr ,-1.)
  call from_param('minr',minr,-1.)
  call from_param('maxr',maxr,-1.)

  ! input pick file parameters
  if (exist_file('hff')) then
    call from_aux('hff','n1',nh1,1 )
    call from_aux('hff','n2',nh2,1 )
  else
    call seperr('Need header history file')
  end if

  ! noise parameters
  call from_param('opt',opt,1)
  call from_param('seed',seed,1)
  call from_param('rms',rms,1.)


  !===================================
  ! initialize
  !===================================
  allocate(model(n1*n2))
  allocate(header(nh1,nh2))


  !===================================
  ! QC/clean out traces
  !===================================
  ! read in header information
  call sreed('hff',header,4*nh1*nh2)
  ! get rid of reciprocal traces
  ! needs to be based on line and point coordinates
  ntraces = nh2
  nkeep = 0
  do ii = 1,ntraces
    ! grab all (line,point) coordinates
    rline  = header(rlkey,ii)
    rpoint = header(rpkey,ii)
    sline  = header(slkey,ii)
    spoint = header(spkey,ii)
    ! take only the lower values
    if (rline.gt.sline) cycle
    if (rline.eq.sline .and. rpoint.gt.spoint) cycle
    nkeep = nkeep+1
    ! move trace up in vector
    header(:,nkeep) = header(:,ii)
  end do
  ! zero out the rest of the vector
  ntraces = nkeep
  if (ntraces.ne.nh2) header(:,ntraces+1:nh2) = 0.
write(0,*) "after non-reciprocal traces:",ntraces

  ! based on snr
  nkeep = 0
  do ii = 1,ntraces
    ! some sort of snr filter
    if (snr.ne.-1 .and. header(snrkey,ii).lt.snr) cycle
    nkeep = nkeep+1
    header(:,nkeep) = header(:,ii)
  end do
  ! zero out the rest of the vector
  ntraces = nkeep
  if (ntraces.ne.nh2) header(:,ntraces+1:nh2) = 0.
write(0,*) "after snr filter:",ntraces

  ! based on min and max distance between receivers
  nkeep = 0
  do ii = 1,ntraces
    ! some sort of distance filter
    dist = sqrt( (header(rxkey,ii)-header(sxkey,ii))**2. + (header(rykey,ii)-header(sykey,ii))**2. )
    if (minr.ne.-1 .and. minr.gt.dist) cycle
    if (maxr.ne.-1 .and. maxr.lt.dist) cycle
    nkeep = nkeep+1
    header(:,nkeep) = header(:,ii)
  end do
  ! zero out the rest of the vector
  ntraces = nkeep
  if (ntraces.ne.nh2) header(:,ntraces+1:nh2) = 0.
write(0,*) "after min/max distance:",ntraces



  !===================================
  ! initialization
  !===================================
  ! allocate synthetic travel times
  allocate(dat(ntraces))

  ! initialize the tomography
  call tomo_straightray_init(ntraces,header(rxkey,:),header(rykey,:),header(sxkey,:),header(sykey,:),n1,o1,d1,n2,o2,d2)



  !===================================
  ! kernel
  !===================================
  ! loop over all models
  do jj=1,n3
    ! read in models and convert to slowness
    call sreed('in',model,4*n1*n2)
    model = 1/model

    ! create travel times
    call tomo_straightray_for(model,dat)

    ! add noise based on rms
    if (opt.eq.2) then
      ! loop over all travel times again
      call random_init(seed)
      do ii=1,ntraces,2
        stdev1 = rms*dat(ii)
        stdev2 = rms*dat(ii+1)
        call random(rand1)
        call random(rand2)
        noise1 = stdev1 * sqrt(-2.*log(rand1)) * cos(2*pi*rand2)
        noise2 = stdev2 * sqrt(-2.*log(rand1)) * sin(2*pi*rand2)
        dat(ii)   = dat(ii)  +noise1
        dat(ii+1) = dat(ii+1)+noise2
      end do

      ! check if ntraces was odd
      if (mod(ntraces,2).ne.0) then
        stdev1 = rms*dat(ntraces)
        call random(rand1)
        call random(rand2)
        noise1 = stdev1 * sqrt(-2.*log(rand1)) * cos(2*pi*rand2)
        dat(ntraces) = dat(ntraces)+noise1
      end if
    end if

    ! create file to output
    allocate(datout(nh1,ntraces))
    datout(rxkey,1:ntraces) = header(rxkey,1:ntraces)
    datout(rykey,1:ntraces) = header(rykey,1:ntraces)
    datout(sxkey,1:ntraces) = header(sxkey,1:ntraces)
    datout(sykey,1:ntraces) = header(sykey,1:ntraces)
    datout(rlkey,1:ntraces) = header(rlkey,1:ntraces)
    datout(rpkey,1:ntraces) = header(rpkey,1:ntraces)
    datout(slkey,1:ntraces) = header(slkey,1:ntraces)
    datout(spkey,1:ntraces) = header(spkey,1:ntraces)
    datout(ttkey,1:ntraces) = dat(1:ntraces)
    datout(snrkey,1:ntraces) = header(snrkey,1:ntraces)

    ! write out data
    call srite('out',datout,4*nh1*ntraces)

  end do


  !===================================
  ! write out
  !===================================
  call to_history('n1',nh1)
  call to_history('n2',ntraces)
  call to_history('o2',0. )
  call to_history('d2',1. )
  call to_history('n3',n3 )


  !===================================
  ! clean up
  !===================================
  deallocate(model,header,dat)
  call sep_close()
  call tomo_straightray_close()
  call exit(0)

end program
