program jcRemoveThreshold
  ! Program to remove all traces above (or below) a certain value
  ! Written expecting threshold parameter to be input
  ! Based solely on the maximum absolute value of the information
  !
  ! Input:
  !   1) history file
  !      - written assuming that the inputs are not traces, but rather
  !      information
  !   2) thresh [0]: parameter which you are basing filter on
  !   3) below:
  !      - [0]: throw out anything above
  !      - [1]: throw out anything below
  !   4) ind
  !      - [0]: no index in particular
  !      - else choose the index value to check
  !   5) header file (hff_in)
  !
  ! Output:
  !   1) history file (probably tapered)
  !      - note: if input has two axes (n2,n3), the output has one axis (n2)
  !   2) header file (hff_out)
  !
  ! Jason Chang
  ! January 2015

  !=================================================
  ! standard declarations
  !================================================
  use sep
  implicit none


  !=================================================
  ! allocate variables
  !================================================
  ! program variables
  integer                       :: ii

  ! history file parameters
  integer                       :: n1,n2,n3
  real                          :: o1,o2,o3,d1,d2,d3
  real,allocatable,dimension(:) :: datain

  ! header file parameters
  integer                       :: n1hdr,n2hdr,n3hdr
  real                          :: o1hdr,o2hdr,o3hdr,d1hdr,d2hdr,d3hdr
  real,allocatable,dimension(:) :: hdrin

  ! function variables
  integer                       :: indx,ntraces
  real                          :: thresh
  logical                       :: below


  !=================================================
  ! open I/O
  !================================================
  call sep_init()

  ! history file parameters
  call from_history('n1',n1,1 )
  call from_history('n2',n2,1 )
  call from_history('n3',n3,1 )
  call from_history('o1',o1,0.)
  call from_history('o2',o2,0.)
  call from_history('o3',o3,0.)
  call from_history('d1',d1,1.)
  call from_history('d2',d2,1.)
  call from_history('d3',d3,1.)

  ! header file parameters
  if (exist_file('hff_in')) then
    call from_aux('hff_in','n1',n1hdr,1 )
    call from_aux('hff_in','n2',n2hdr,1 )
    call from_aux('hff_in','n3',n3hdr,1 )
    call from_aux('hff_in','o1',o1hdr,0.)
    call from_aux('hff_in','o2',o2hdr,0.)
    call from_aux('hff_in','o3',o3hdr,0.)
    call from_aux('hff_in','d1',d1hdr,1.)
    call from_aux('hff_in','d2',d2hdr,1.)
    call from_aux('hff_in','d3',d3hdr,1.)
    if ((n2hdr*n3hdr).ne.(n2*n3)) call seperr('Second dimensions do not match')
  end if

  ! input parameters
  call from_param('thresh',thresh,0.)
  call from_param('below' ,below ,.false.)
  call from_param('indx'  ,indx  ,0)


  !=================================================
  ! initialize
  !================================================
  ! allocate
  allocate(datain(n1))
  allocate(hdrin(n1hdr))

  ! initialize counter
  ntraces = 0


  !=================================================
  ! kernel
  !================================================
  do ii=1,n2*n3
    ! read in data
    call sreed('in',datain,4*n1)
    if (exist_file('hff_in')) call sreed('hff_in',hdrin,4*n1hdr)

    ! grab correct index
    if (indx.eq.0) then
      if (below) then
        if (thresh.lt.maxval(datain(:))) then
          call srite('out',datain,4*n1)
          if (exist_file('hff_out')) call srite('hff_out',hdrin,4*n1hdr)
          ntraces = ntraces+1
        end if
      else
        if (thresh.ge.maxval(datain(:))) then
          call srite('out',datain,4*n1)
          if (exist_file('hff_out')) call srite('hff_out',hdrin,4*n1hdr)
          ntraces = ntraces+1
        end if
      end if
    else
      if (below) then
        if (thresh.lt.datain(indx)) then
          call srite('out',datain,4*n1)
          if (exist_file('hff_out')) call srite('hff_out',hdrin,4*n1hdr)
          ntraces = ntraces+1
        end if
      else
        if (thresh.ge.datain(indx)) then
          call srite('out',datain,4*n1)
          if (exist_file('hff_out')) call srite('hff_out',hdrin,4*n1hdr)
          ntraces = ntraces+1
        end if
      end if
    end if

  end do

  ! status update
  write(0,*) "Percentage of traces kept:",real(ntraces)/(n2*n3)


  !=================================================
  ! write out parameters
  !=================================================
  call to_history('n1',n1)
  call to_history('n2',ntraces)
  call to_history('n3',1 )
  call to_history('o1',o1)
  call to_history('o2',o2)
  call to_history('o3',0.)
  call to_history('d1',d1)
  call to_history('d2',d2)
  call to_history('d3',1.)

  if (exist_file('hff_out')) then
    call to_history('n1',n1hdr  ,'hff_out')
    call to_history('n2',ntraces,'hff_out')
    call to_history('n3',1      ,'hff_out')
    call to_history('o1',o1hdr  ,'hff_out')
    call to_history('o2',o2hdr  ,'hff_out')
    call to_history('o3',0.     ,'hff_out')
    call to_history('d1',d1hdr  ,'hff_out')
    call to_history('d2',d2hdr  ,'hff_out')
    call to_history('d3',1.     ,'hff_out')
  end if


  !=================================================
  ! clean up
  !================================================
  call sep_close()
  deallocate(datain,hdrin)
  call exit(0)


end program
