program jcTravelTimePick_LB
  !=============================================================================================
  ! Program to perform group travel time picking
  ! Modified for Long Beach data (November 2014)
  !   - now includes receiver and source line index and number index
  ! Modified to mute out predicted region of higher-order mode (February 2015)
  !   - now includes v,t0 for higher-order mode
  ! Modified to include an event prediction window (February 2015)
  !   - because highway energy does not have a zero-orgin time
  !
  ! Inputs:
  !	1) history file correlations (negative and positive time lags)
  ! 		- n1: number of time samples
  !		- n2: number of correlations (can also extend up to n6)
  !	2) hff: header history file correlations
  !     3) sym: whether to symmetrize inputs
  !             - sym_out: if you want the output symmetry file
  !     4) bandpass: whether to apply bandpass filter or not
  !             - ncfreq,ocfreq,dcfreq,alpha (scaling parameter of sorts)
  !             - bp_out: if you want the output bandpassed file
  !     5) env_out: if you want the output enveloped trace
  !     6) snr: whether you want to calculate snr
  !             - tau: how many seconds on each side of max amplitude
  !             - recommended only for symmetrized traces
  !     7) mute: for muting higher-mode
  !             - vm1,tm1: slower velocity and starting time for predicting higher-mode window
  !             - vm2,tm2: faster velocity and starting time for predicting higher-mode window
  !             - these are obtained via FK domain observations
  !             - make sure the units make sense
  !             - recommended only for symmetrized traces
  !     8) wind: for prediction window of arrival
  !             - vp1,tp1: slower velocity and starting time for predicting fundamental-mode window
  !             - vp2,tp2: faster velocity and starting time for predicting fundamental-mode window
  !             - these are obtained via FK domain observations
  !             - make sure the units make sense
  !             - recommended only for symmetrized traces
  !
  !
  ! Outputs:
  !	1) history file
  !		- n1: 5 to 6 (travel time, sourcex, sourcey, receiverx, receivery, snr)
  !		- n2: number of correlations
  !		- n3: number of frequencies
  !
  ! Order of operations:
  !     0) indices of rx,ry,sx,sy,rline,rpoint,sline,spoint
  !		- rxind: rx hdr index [21]
  !		- ryind: ry hdr index [22]
  !		- sxind: sx hdr index [23]
  !		- syind: sy hdr index [24]
  !             - rlind: rline index  [3]
  !             - rpind: rpoint index [4]
  !             - slind: sline index  [7]
  !             - spind: spoint index [8]
  !     1) taper in time
  !		- taper: whether to taper or not
  !			  0. do not taper
  !			[1]. taper
  !		- topt: taper options
  !		          0. hann-half filter [t1,t2]
  !                       1. hann filter [t1,t2]
  !                     [2]. hann-band filter [t1,t2,t3,t4]
  !		- taper_out: name of file for output of taper
  !	2) bandpass with gaussian filter
  !		- bandpass: whether to bandpass or not
  !			  0. do not bandpass
  !			[1]. bandpass
  !		- fopt: bandpass filter options
  !			  0. heaviside (for analytic function)
  !			[1]. gaussian [ncfreq,ocfreq,dcfreq,alpha]
  !			  2. gaussian heaviside [ncfreq,ocfreq,dcfreq,alpha--controls width of gaussian]
  !			  3. hann [ncfreq,ocfreq,dcfreq,alpha--defines boundaries of filter around cfreq] 
  !		- bp_out: name of file for output of bandpassed signal
  !     3) symmetrize
  !		- sym: whether to symmetrize trace or not
  !			[0]. do not symmetrize
  !			  1. symmetrize
  !		- sym_out: name of symmetrized file
  !  	4) SNR filter
  !		- snr: whether to output snr or not
  !			  0. do not output snr
  !                     [1]. output snr
  !		- tau: seconds on each side of max amplitude [1.]
  !	5) calculate envelope
  !		- env_out: name of file for output of enveloped signal
  !	6) compute travel time (perhaps based on picking window)
  !
  !
  ! Jason Chang
  ! April 2014
  !=============================================================================================

  !==========================================================
  ! includes
  !==========================================================
  use sep
  use symmetrize_mod
  use taper_mod
  use bandpass_mod
  use snr_mod
  implicit none

  !==========================================================
  ! variables
  !==========================================================
  ! program variables
  integer                                 :: ii,jj,kk
  real,allocatable,dimension (:)          :: dat
  complex,allocatable,dimension (:)       :: cdat,cdatin,cdatsym

  ! I/O variables
  integer                                 :: n1,n2
  integer                                 :: n1xy,n2xy
  integer                                 :: rxind,ryind,sxind,syind
  integer                                 :: rlind,rpind,slind,spind
  real                                    :: d1,d2,o1,o2
  real,allocatable,dimension (:)          :: datin,hdrin

  ! symmetrize variables
  integer                                 :: nt
  real                                    :: ot
  complex,allocatable,dimension (:)       :: temp
  logical                                 :: sym

  ! taper variables
  integer                                 :: topt
  real                                    :: t1,t2,t3,t4
  logical                                 :: taper

  ! bandpass variables
  integer                                 :: fopt
  integer                                 :: ncfreq
  real                                    :: alpha,ocfreq,dcfreq,cfreq
  logical                                 :: bandpass

  ! snr variables
  integer                                 :: tau_ind
  real                                    :: thresh,tau,vsnr
  logical                                 :: snr

  ! mute variables
  integer                                 :: itm1,itm2
  real                                    :: vm1,vm2,tm1,tm2,ttm1,ttm2
  real                                    :: dist,dist1,dist2
  logical                                 :: mute

  ! prediction window variables
  integer                                 :: itp1,itp2
  real                                    :: vp1,vp2,tp1,tp2,ttp1,ttp2
  logical                                 :: wind

  ! travel time variables
  integer                                 :: itt
  real                                    :: tt


  !==========================================================
  ! open I/O
  !==========================================================
  call sep_init()

  ! history file parameters
  call from_history('n1',n1,1 )
  call from_history('o1',o1,0.)
  call from_history('d1',d1,1.)
  call from_history('n2',n2,1 )
  call from_history('o2',o2,0.)
  call from_history('d2',d2,1.)

  ! header history file parameters
  call from_param('rxind',rxind,21)
  call from_param('ryind',ryind,22)
  call from_param('sxind',sxind,23)
  call from_param('syind',syind,24)
  call from_param('rlind',rlind,3 )
  call from_param('rpind',rpind,4 )
  call from_param('slind',slind,7 )
  call from_param('spind',spind,8 )
  if (exist_file('hff')) then
    call from_aux('hff','n1',n1xy)
    call from_aux('hff','n2',n2xy)
    if ( n2 .ne. n2xy ) call erexit ("n2 and n2xy do not match")
  else
    call erexit('Need a header file')
  end if

  ! taper parameters
  call from_param('taper',taper,.true.)
  call from_param('topt',topt,2)
  call from_param('t1',t1,o1)
  if (topt.eq.2) then
    call from_param("t2",t2,o1+1*n1/3*d1)
  else
    call from_param("t2",t2,o1+(n1-1)*d1)
  end if
  call from_param ("t3",t3,o1+2*n1/3*d1)
  call from_param ("t4",t4,o1+(n1-1)*d1)
  if (exist_file('taper_out')) then
    call to_history('n1',n1,'taper_out')
    call to_history('o1',o1,'taper_out')
    call to_history('d1',d1,'taper_out')
    call to_history('n2',n2,'taper_out')
    call to_history('o2',o2,'taper_out')
    call to_history('d2',d2,'taper_out')
  end if


  ! bandpass parameters
  call from_param('bandpass',bandpass,.true.)
  call from_param('fopt',fopt,1)
  call from_param('alpha',alpha,10.)
  if (bandpass) then
    call from_param('ncfreq',ncfreq,1 )
    call from_param('ocfreq',ocfreq,cfreq)
    call from_param('dcfreq',dcfreq,1.)
    if (exist_file('bp_out')) then
      call to_history('n1',n1,'bp_out')
      call to_history('o1',o1,'bp_out')
      call to_history('d1',d1,'bp_out')
      call to_history('n2',ncfreq,'bp_out')
      call to_history('o2',ocfreq,'bp_out')
      call to_history('d2',dcfreq,'bp_out')
      call to_history('n3',n2,'bp_out')
      call to_history('o3',o2,'bp_out')
      call to_history('d3',d2,'bp_out')
    end if
  else
      ncfreq = 1
      ocfreq = 0.
      dcfreq = 1.
  end if

  ! symmetrize parameters
  call from_param('sym',sym,.false.)
  if (sym) then
    nt = n1/2+1
    ot = 0.
    if (exist_file('sym_out')) then
      call to_history('n1',nt,'sym_out')
      call to_history('o1',ot,'sym_out')
      call to_history('d1',d1,'sym_out')
      call to_history('n2',ncfreq,'sym_out')
      call to_history('o2',ocfreq,'sym_out')
      call to_history('d2',dcfreq,'sym_out')
      call to_history('n3',n2,'sym_out')
      call to_history('o3',o2,'sym_out')
      call to_history('d3',d2,'sym_out')
    end if
  else
    nt = n1
    ot = o1
  end if

  ! snr parameters
  call from_param('snr',snr,.true.)
  call from_param('thresh',thresh,1.)
  call from_param('tau',tau,1.)
  call from_param('tau_ind',tau_ind,nint((tau/d1)+1))

  ! envelope parameters
  if (exist_file('env_out')) then
    call to_history('n1',nt,'env_out')
    call to_history('o1',ot,'env_out')
    call to_history('d1',d1,'env_out')
    call to_history('n2',ncfreq,'env_out')
    call to_history('o2',ocfreq,'env_out')
    call to_history('d2',dcfreq,'env_out')
    call to_history('n3',n2,'env_out')
    call to_history('o3',o2,'env_out')
    call to_history('d3',d2,'env_out')
  end if

  ! higher-mode mute filter parameters
  call from_param('mute',mute,.false.)
  call from_param('vm1',vm1,0.)
  call from_param('vm2',vm2,0.)
  call from_param('tm1',tm1,0.)
  call from_param('tm2',tm2,0.)

  ! fundamental mode prediction window parameters
  call from_param('wind',wind,.false.)
  call from_param('vp1',vp1,0.)
  call from_param('vp2',vp2,0.)
  call from_param('tp1',tp1,ot)
  call from_param('tp2',tp2,ot+(nt-1)/d1)


  !==========================================================
  ! allocate
  !==========================================================
  ! history file
  allocate(dat(nt))
  allocate(datin(n1))
  allocate(cdatsym(nt))
  allocate(cdatin(n1))
  allocate(cdat(n1))
  allocate(temp(n1))

  ! header history file
  allocate(hdrin(n1xy))


  !==========================================================
  ! initializations
  !==========================================================
  call taper_init(n1,o1,d1,topt)
  call bandpass_init(n1,o1,d1,fopt)
  call symmetrize_init(n1)
  call snr_init(nt,thresh,tau_ind)


  !==========================================================
  ! kernel
  !==========================================================
  do jj=1,n2 ! loop over all traces

    !============================
    ! read in trace
    !============================
    call sreed('in',datin,4*n1)
    call sreed('hff',hdrin,4*n1xy)

    !============================
    ! apply taper
    !============================
    call apply_taper(datin,t1,t2,t3,t4)
    if (exist_file('taper_out')) call srite('taper_out',datin,4*n1)
    do kk=1,n1 ! convert from real to complex
      cdatin(kk) = cmplx(datin(kk),0.)
    end do

    ! fftshift
    temp = cdatin
    cdatin(n1-(n1+1)/2+1:n1) = temp(1:(n1+1)/2)
    cdatin(1:n1-(n1+1)/2) = temp((n1+1)/2+1:n1)


    !============================
    ! loop over all central frequencies
    !============================
    do ii=1,ncfreq
      cfreq = ocfreq+(ii-1)*dcfreq

      !=============================
      ! set complex trace
      !=============================
      cdat = cdatin

      !=============================
      ! bandpass
      !=============================
      if (bandpass) then
        call apply_bandpass(cdat,cfreq,alpha)
        !call gaussian_heaviside_filter(cdat,cfreq,alpha)
        !call hann_filter(cdat,0.5,1.)
      else
        call heaviside_filter(cdat)
      end if
      ! ifftshift
      temp = cdat
      cdat(1:(n1+1)/2) = temp(n1-(n1+1)/2+1:n1)
      cdat((n1+1)/2+1:n1) = temp(1:n1-(n1+1)/2)
      if (exist_file('bp_out')) call srite('bp_out',real(cdat),4*n1)


      !==============================
      ! symmetrize option
      !==============================
      if (sym) then
        call symmetrize_fold_cmplx(cdatsym,cdat)
        if (exist_file('sym_out')) call srite('sym_out',real(cdatsym),4*nt)
      else
        cdatsym = cdat
      end if


      !==============================
      ! mute option
      !==============================
      if (mute.or.wind) then
        ! calculate distance between source and receiver
        dist1 = hdrin(rxind)-hdrin(sxind)
        dist2 = hdrin(ryind)-hdrin(syind)
        dist  = sqrt((dist1**2)+(dist2**2))
      end if
        
      if (mute) then
        ! calculate times
        ttm1 = tm2+(dist/vm2)
        ttm2 = tm1+(dist/vm1)

        ! convert to indices
        itm1 = 1.5+(ttm1-ot)/d1
        itm2 = 1.5+(ttm2-ot)/d1

        ! check the values don't exceed limits
        itm1 = max(1,itm1)
        itm2 = min(nt,itm2)

        ! apply simple mute
        do kk = itm1,itm2
          cdatsym(kk) = 0.
        end do
      end if


      !==============================
      ! predicted window option
      !==============================
      if (wind) then
        ! calculate times
        ttp1 = tp2+(dist/vp2)
        ttp2 = tp1+(dist/vp1)

        ! convert to indices
        itp1 = 1.5+(ttp1-ot)/d1
        itp2 = 1.5+(ttp2-ot)/d1

        ! check the values don't exceed limits
        itp1 = max(1,itp1)
        itp2 = min(nt,itp2)
      end if


      !=============================
      ! take real part and calculate SNR
      ! currently relies on maximum amplitude location
      !=============================
      if (snr) then
        do kk=1,nt
          dat(kk) = real(cdatsym(kk))
        end do
        !vsnr = calc_snr_max(dat)
        !vsnr = calc_snr_max_mute(dat,itm1,itm2)
        vsnr = calc_snr_max_mute_window(dat,itm1,itm2,itp1,itp2)
      end if


      !=============================
      ! take envelope
      !=============================
      do kk=1,nt
        dat(kk) = cabs(cdatsym(kk))
      end do
      if (exist_file('env_out')) call srite('env_out',dat,4*nt)


      !=============================
      ! calculate arrival time (maximum of amplitude)
      !=============================
      tt = abs(dat(itp1))
      itt = itp1
      do kk=itp1+1,itp2
        if (abs(dat(kk)).gt.tt) then
          itt = kk
          tt = abs(dat(kk))
        end if
      end do
      tt = ot + (itt-1)*d1


      !=============================
      ! write out file: RX,RY,SX,SY,TT,RLINE,RPOINT,SLINE,SPOINT(,SNR)
      !=============================
      call srite('out',hdrin(rxind),4)
      call srite('out',hdrin(ryind),4)
      call srite('out',hdrin(sxind),4)
      call srite('out',hdrin(syind),4)
      call srite('out',hdrin(rlind),4)
      call srite('out',hdrin(rpind),4)
      call srite('out',hdrin(slind),4)
      call srite('out',hdrin(spind),4)
      call srite('out',tt,4)
      if (snr) call srite('out',vsnr,4)

    end do
  end do


  !==========================================================
  ! write out parameters
  !==========================================================
  if (snr) call to_history('n1',10)
  if (.not.snr) call to_history('n1',9)
  call to_history('o1',1.)
  call to_history('d1',1.)
  call to_history('n2',ncfreq)
  call to_history('o2',ocfreq)
  call to_history('d2',dcfreq)
  call to_history('n3',n2)
  call to_history('o3',o2)
  call to_history('d3',d2)


  !==========================================================
  ! clean up
  !==========================================================
  deallocate(cdatsym,datin,cdat,cdatin,temp,dat)
  call taper_close()
  call bandpass_close()
  call symmetrize_close()
  call snr_close()
  call sep_close()
  call exit(0)

end program
