module laplacian2dOMP_mod
  ! Module to calculate laplacian
  ! Primarily for regularization
  ! In GEE: pgs 62,78
  !
  ! Jason Chang
  ! November 2014

  !=======================================
  ! includes
  !=======================================
  use adj_mod
  implicit none


  !=======================================
  ! declare module variables
  !=======================================
  integer,private                       :: ngx,ngy
  real,allocatable,dimension(:),private :: temp

  contains


  !=======================================
  ! initialization subroutine
  !=======================================
  subroutine laplacian2dOMP_init(ngx_in,ngy_in)
    ! declare variables
    integer           :: ngx_in,ngy_in

    ! pass module variables
    ngx = ngx_in
    ngy = ngy_in

    ! allocate
    allocate(temp(ngx*ngy*2))

  end subroutine


  !=======================================
  ! clean up subroutine
  !=======================================
  subroutine laplacian2dOMP_close()
    deallocate(temp)
  end subroutine


  !=======================================
  ! operator for SEP solvers
  !=======================================
  function laplacian2dOMP_lop(adj,add,m,d) result(stat)
    ! declare variables
    integer            :: stat
    logical,intent(in) :: adj,add
    real,dimension(:)  :: m,d

    ! begin calling functions
    call adjnull (adj,add,m,d)
    if (.not.adj) call laplacian2dOMP_for(m,d)
    if (     adj) call laplacian2dOMP_adj(m,d)
    stat = 0.
  end function


  !=======================================
  ! forward laplacian subroutine
  !=======================================
  subroutine laplacian2dOMP_for(m,d)
    ! declare variables
    integer                       :: stat1,stat2
    real,dimension(:),intent(in)  :: m
    real,dimension(:),intent(out) :: d

    ! call gradients twice
    stat1 = gradient2dOMP_lop(.false.,.false.,m,temp)
    stat2 = gradient2dOMP_lop(.true. ,.true. ,d,temp)

  end subroutine


  !=======================================
  ! adjoint laplacian subroutine
  !=======================================
  subroutine laplacian2dOMP_adj(m,d)
    ! declare variables
    integer                       :: stat1,stat2
    real,dimension(:),intent(in)  :: d
    real,dimension(:),intent(out) :: m

    ! call gradients twice
    stat1 = gradient2dOMP_lop(.false.,.false.,d,temp)
    stat2 = gradient2dOMP_lop(.true. ,.true. ,m,temp)
    
  end subroutine


  !=======================================
  ! operator for SEP solvers
  !=======================================
  function gradient2dOMP_lop(adj,add,m,d) result(stat)
    ! declare variables
    integer                      :: stat
    real,dimension(ngx,ngy)      :: m
    real,dimension(ngx,ngy,2)    :: d
    logical                      :: adj,add

    ! begin calling functions
    call adjnull2(adj,add,m,d)
    if (.not.adj) call gradient2dOMP_for(m,d)
    if (     adj) call gradient2dOMP_adj(m,d)
    stat = 0.
  end function


  !=======================================
  ! forward gradient
  !=======================================
  subroutine gradient2dOMP_for(m,d)
    ! declare variables
    integer                           :: ii,jj
    real,dimension(:,:),intent(in)    :: m
    real,dimension(:,:,:),intent(out) :: d

    ! kernel
    !$OMP PARALLEL DO DEFAULT(PRIVATE) SHARED(d,m,ngx,ngy)
    do ii = 1,ngx-1
      do jj = 1,ngy-1
        d(ii,jj,1) = d(ii,jj,1)+m(ii+1,jj)-m(ii,jj)
        d(ii,jj,2) = d(ii,jj,2)+m(ii,jj+1)-m(ii,jj)
      end do
    end do
    !$OMP END PARALLEL DO

   end subroutine


  !=======================================
  ! adjoint gradient
  !=======================================
  subroutine gradient2dOMP_adj(m,d)
    ! declare variables
    integer                          :: ii,jj
    real,dimension(:,:,:),intent(in) :: d
    real,dimension(:,:),intent(out)  :: m

    ! kernel
    !$OMP PARALLEL DO DEFAULT(PRIVATE) SHARED(d,m,ngx,ngy)
    do ii = 1,ngx-1
      do jj = 1,ngy-1
        !$OMP ATOMIC
        m(ii+1,jj  ) = m(ii+1,jj)+d(ii,jj,1)
        !$OMP ATOMIC
        m(ii  ,jj  ) = m(ii,jj)-d(ii,jj,1)-d(ii,jj,2)
        !$OMP ATOMIC
        m(ii  ,jj+1) = m(ii,jj+1)+d(ii,jj,2)
      end do
    end do
    !$OMP END PARALLEL DO
   
  end subroutine


  !=======================================
  ! special adj_null
  !=======================================
  subroutine     adjnull2(adj,add,x,y)
    logical,intent(in) :: adj , add
    real               :: x(:,:), y(:,:,:)
    if (.not.add) then
      if (     adj)  x = 0.
      if (.not.adj)  y = 0.
    end if
  end subroutine

end module
