module regsolver_mod
  ! Module for solver with regularization
  ! GEE: pg 84
  !
  ! Jason Chang
  ! November 2014

  !======================================
  ! standard declarations
  !======================================
  use chain_mod
  use solver_report_mod
  implicit none


  !======================================
  ! module variables
  !======================================
  logical,parameter,private   :: T=.true.,F=.false.

  contains


  subroutine regsolver(m,d,Fop,Aop,stepper,nAop,niter,eps,Wop,m0,rm0,err,resd,resm,mmov,rmov,verb)
    optional                                           :: Wop,m0,rm0,err,resd,resm,mmov,rmov,verb

    ! description of variables
    ! m: model
    ! d: data
    ! Fop: data-fitting operator
    ! Aopt: regularization operator
    ! stepper: step operator
    ! nAop: size of regularizaton output
    ! niter: number of iterations
    ! eps: regularization weight
    ! Wop: weighting operator
    ! m0: initial model
    ! rm0: iniital model residual
    ! err: norm of errors
    ! resd: final data residual value
    ! resm: final model residual values
    ! mmov: model outputs
    ! rmov: residual outputs
    ! verb: verbose
     

    interface
      ! forward operator
      integer function Fop(adj,add,m,d)
        logical,intent(in) :: adj,add
        real,dimension(:)  :: m,d
      end function

      ! regularization operator
      integer function Aop(adj,add,m,d)
        logical,intent(in) :: adj,add
        real,dimension(:)  :: m,d
      end function
  
      ! weighting operator
      integer function Wop(adj,add,m,d)
        logical,intent(in) :: adj,add
        real,dimension(:)  :: m,d
      end function

      ! step function
      integer function stepper(forget,m,dm,r,dr)
        real,dimension(:)  :: m,dm,r,dr
        logical            :: forget
      end function

    end interface

    ! declare variables
    integer                             :: iter,stat
    integer,intent(in)                  :: niter,nAop
    real,intent(in)                     :: eps
    real,dimension(:),intent(in)        :: d,m0,rm0
    logical,intent(in)                  :: verb
    real,dimension(:),intent(out)       :: m,resd,resm
    real,dimension(:,:),intent(out)     :: rmov,mmov,err
    real,dimension(size(m))             :: dm
    real,dimension(size(d)+nAop),target :: r,dr,tt
    real,dimension(:),pointer           :: rd,drd,td
    real,dimension(:),pointer           :: rm,drm,tm
    logical                             :: forget

    ! dm is model perturbation (delta m)
    ! rd is data-space residual
    ! rm is model-space residual
    ! drd is the change in data-space residual
    ! drm is the change in model-space residual

    ! direct pointers
    rd  => r(1:size(d))
    rm  => r(1+size(d):)
    drd => dr(1:size(d))
    drm => dr(1+size(d):)
    td  => tt(1:size(d))
    tm  => tt(1+size(d):)

    ! zero the model and the model residual
    m = 0.
    rm = 0.

    ! initialization
    if (present(rm0)) rm = rm0        ! Rm = Rm0
    if (present(m0))  m  = m0         ! m  = m0
    if (present(Wop)) then
      stat = Wop(F,F,-d,rd)           ! Rd = W * -d
      call chain2(Wop,Fop,F,T,m,rd,td)  ! Rd = Rd + W*F*m0
    else
      rd   = -d                       ! Rd = -d
      stat = Fop(F,T,m,rd)            ! Rd = Rd + F*m0 (or Fm0 - d)
    end if
    stat = Aop(F,T,eps*m,rm)          ! Rm = Rm + e*A*m0

    forget = T

    ! begin iterations
    do iter = 1,niter

      ! compute proposed model perturbation
      if (present(Wop)) then
        call chain2(Wop,Fop,T,F,dm,rd,td) ! dm = (W*F)' * rd
      else
        stat = Fop(T,F,dm,rd)             ! dm = F' * rd
      end if
      stat = Aop(T,T,dm,eps*rm)           ! dm = dm + e*A'*rm

      ! compute change in data residual and change in model residual
      if (present(Wop)) then
        call chain2(Wop,Fop,F,F,dm,drd,td) ! drd = W*F*dm
      else
        stat = Fop(F,F,dm,drd)             ! drd = F*dm
      end if
      stat = Aop(F,F,eps*dm,drm)           ! drm = eps*A*dm

      ! step
      stat = stepper(forget,m,dm,r,dr) ! m = m+dm; r = r+dr
      if (stat.eq.1) exit ! got stuck descending

      if (present( mmov)) mmov(:,iter) =  m(:size(mmov,1))
      if (present( rmov)) rmov(:,iter) =  r(:size(rmov,1))
      if (present( err)) then
        err(iter,1) = dot_product(rd,rd)    ! data residual
        err(iter,2) = dot_product(drd,drd)  ! change in data residual
        err(iter,3) = dot_product(rm,rm)    ! model residual
        err(iter,4) = dot_product(drm,drm)  ! change in model residual
      end if
      if(verb) write(0,*) "iter",iter,err(iter,1),err(iter,2),err(iter,3),err(iter,4)
      forget=F            ! Steepest descent ?
      
    end do

    ! write out final residuals
    if (present( resd)) resd = rd
    if (present( resm)) resm = rm(:size(resm))

  end subroutine

end module
