module snr_mod
  ! Module containing snr filter
  !
  ! Jason Chang
  ! April 2014

  !==========================================================
  ! standard declarations
  !==========================================================
  implicit none

  !==========================================================
  ! declare module variables
  !==========================================================
  integer,private                        :: n1,tau_ind
  real,private                           :: thresh

  contains

  !==========================================================
  ! snr initialization initialization
  !==========================================================
  subroutine snr_init(n1_in,thresh_in,tau_ind_in)

    !===========================================
    ! declare variables
    !===========================================
    integer                      :: n1_in,tau_ind_in
    real                         :: thresh_in

    !===========================================
    ! set these variables for entire module
    !===========================================
    n1 = n1_in
    thresh = thresh_in
    tau_ind = tau_ind_in

  end subroutine


  !=============================================================
  ! close subroutine
  !=============================================================
  subroutine snr_close()
  end subroutine


  !=============================================================
  ! return logical variable for snr based on time of maximum amplitude
  !=============================================================
  function logical_snr_max(dat) result(vsnr)

    !=============================
    ! declare variables
    !=============================
    integer                        :: ii,counter
    integer                        :: maxamploc,wlo,whi
    real                           :: maxamp,noise,snr
    real,dimension (:)             :: dat
    logical                        :: vsnr

    !=============================
    ! reset values
    !=============================
    maxamp = 0.
    counter = 0
    noise = 0.
    snr = 0.

    !=============================
    ! find the index and value of the maximum absolute amplitude
    !=============================
    do ii=1,n1
      if (abs(dat(ii)) > maxamp) then
        maxamploc = ii
        maxamp = abs(dat(ii))
      end if
    end do
    
    !=============================
    ! define a window centered about maximum amplitude
    !=============================
    ! calculate the centered window around the maximum
    wlo = maxamploc-tau_ind
    whi = maxamploc+tau_ind

    !=============================
    ! check the boundary of the windows
    ! need to try different cases
    !=============================
    ! if window is too big, then do nothing
    if ((wlo .lt. 1) .and. (whi .gt. n1)) then
      call erexit('Window is too big--no noise region')
    ! if window meets lower boundary, then only take later than window
    elseif (wlo .lt. 1) then
      do ii=whi+1,n1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
    ! if at higher boundary, then only take earlier than window
    elseif (whi .gt. n1) then
      do ii=1,wlo-1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
    ! otherwise take both earlier than later windows
    else
      do ii=1,wlo-1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
      do ii=whi+1,n1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
    end if

    ! calculate RMS of the noise
    noise = sqrt(noise/counter)
    
    ! compare ratio of max amplitude to RMS noise amplitude
    snr = maxamp/noise
write(0,*) snr

    ! return proper result
    if (snr.lt.thresh) vsnr = .false.
    if (snr.ge.thresh) vsnr = .true.

  end function


  !=============================================================
  ! return snr based on time of maximum amplitude
  !=============================================================
  function calc_snr_max(dat) result(snr)

    !=============================
    ! declare variables
    !=============================
    integer                        :: ii,counter
    integer                        :: maxamploc,wlo,whi
    real                           :: maxamp,noise,snr
    real,dimension (:)             :: dat

    !=============================
    ! reset values
    !=============================
    maxamp = 0.
    counter = 0
    noise = 0.
    snr = 0.

    !=============================
    ! find the index and value of the maximum absolute amplitude
    !=============================
    do ii=1,n1
      if (abs(dat(ii)) > maxamp) then
        maxamploc = ii
        maxamp = abs(dat(ii))
      end if
    end do
    
    !=============================
    ! define a window centered about maximum amplitude
    !=============================
    ! calculate the centered window around the maximum
    wlo = maxamploc-tau_ind
    whi = maxamploc+tau_ind

    !=============================
    ! check the boundary of the windows
    ! need to try different cases
    !=============================
    ! if window is too big, then do nothing
    if ((wlo .lt. 1) .and. (whi .gt. n1)) then
      call erexit('Window is too big--no noise region')
    ! if window meets lower boundary, then only take later than window
    elseif (wlo .lt. 1) then
      do ii=whi+1,n1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
    ! if at higher boundary, then only take earlier than window
    elseif (whi .gt. n1) then
      do ii=1,wlo-1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
    ! otherwise take both earlier than later windows
    else
      do ii=1,wlo-1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
      do ii=whi+1,n1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
    end if

    ! calculate RMS of the noise
    noise = sqrt(noise/counter)
    
    ! compare ratio of max amplitude to RMS noise amplitude
    snr = maxamp/noise

  end function


  
  !=============================================================
  ! return snr based on time of maximum amplitude
  !   - inputs include window of zeros (for zeroed out regions)
  !=============================================================
  function calc_snr_max_mute(dat,ilo,ihi) result(snr) 
    !=============================
    ! declare variables
    !=============================
    integer                        :: ii,counter
    integer                        :: ii1,ii2,ilo,ihi,iL,iH
    integer                        :: maxamploc,wlo,whi
    real                           :: maxamp,noise,snr
    real,dimension (:)             :: dat

    !=============================
    ! reset values
    !=============================
    maxamp = 0.
    counter = 0
    noise = 0.
    snr = 0.

    !=============================
    ! find the index and value of the maximum absolute amplitude
    !=============================
    do ii=1,n1
      if (abs(dat(ii)) > maxamp) then
        maxamploc = ii
        maxamp = abs(dat(ii))
      end if
    end do

    !=============================
    ! define a window centered about maximum amplitude
    !=============================
    ! calculate the centered window around the maximum
    wlo = maxamploc-tau_ind
    whi = maxamploc+tau_ind

    !=============================
    ! check the boundary of the windows
    ! need to try different cases
    !=============================
    ! if window is too big, then do nothing
    if ((wlo .lt. 1) .and. (whi .gt. n1)) then
      call erexit('Window is too big--no noise region')
    end if
    ! set the indices of the region around maximum
    ii1 = max(wlo,1)
    ii2 = min(whi,n1)

    ! if gap is earlier than tau region
    if (ii2.lt.ilo) then
      do ii=1,ii1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
      do ii=ii2,ilo
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
      do ii=ihi,n1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
    ! if gap is later than tau region
    elseif (ihi.lt.ii1) then
      do ii=1,ilo
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
      do ii=ihi,ii1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
      do ii=ii2,n1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
    ! all other situations
    else
      iL = min(ii1,ilo)
      iH = max(ii2,ihi)
      do ii=1,iL
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
      do ii=iH,n1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
    end if

    ! calculate RMS of the noise
    noise = sqrt(noise/counter)

    ! compare ratio of max amplitude to RMS noise amplitude
    snr = maxamp/noise

  end function


  !=============================================================
  ! return snr based on time of maximum amplitude
  !   - inputs include:
  !      - window of zeros (for zeroed out regions)
  !      - window of predicted arrival
  !=============================================================
  function calc_snr_max_mute_window(dat,im1,im2,ip1,ip2) result(snr) 
    !=============================
    ! declare variables
    !=============================
    integer                        :: ii,counter
    integer                        :: im1,im2,ip1,ip2,iL,iH
    integer                        :: maxamploc,wlo,whi
    real                           :: maxamp,noise,snr
    real,dimension (:)             :: dat

    !=============================
    ! reset values
    !=============================
    maxamp = 0.
    counter = 0
    noise = 0.
    snr = 0.

    !=============================
    ! find the index and value of the maximum absolute amplitude
    !=============================
    ! set the indices
    ip1 = max(ip1,1)
    ip2 = min(ip2,n1)
    do ii=ip1,ip2
      if (abs(dat(ii)) > maxamp) then
        maxamploc = ii
        maxamp = abs(dat(ii))
      end if
    end do

    !=============================
    ! define a window centered about maximum amplitude
    !=============================
    ! calculate the centered window around the maximum
    wlo = maxamploc-tau_ind
    whi = maxamploc+tau_ind

    !=============================
    ! check the boundary of the windows
    ! need to try different cases
    !=============================
    ! if window is too big, then do nothing
    if ((wlo .lt. 1) .and. (whi .gt. n1)) then
      call erexit('Window is too big--no noise region')
    end if
    ! set the indices of the region around maximum
    wlo = max(wlo,1)
    whi = min(whi,n1)

    ! if gap is later than tau region
    if (whi.lt.im1) then
      do ii=1,wlo
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
      do ii=whi,im1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
      do ii=im2,n1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
    ! if gap is earlier than tau region
    elseif (im2.lt.wlo) then
      do ii=1,im1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
      do ii=im2,wlo
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
      do ii=whi,n1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
    ! all other situations
    else
      iL = min(wlo,im1)
      iH = max(whi,im2)
      do ii=1,iL
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
      do ii=iH,n1
        noise = noise + dat(ii)**2
        counter = counter + 1
      end do
    end if

    ! calculate RMS of the noise
    noise = sqrt(noise/counter)

    ! compare ratio of max amplitude to RMS noise amplitude
    snr = maxamp/noise

  end function


end module
