module tomo_straightray_mod
  ! Module to perform straight ray tomography
  ! Made to fix old erors (described in test1-mod)
  !
  ! Jason Chang
  ! November 2014

  !=========================================
  ! includes
  !=========================================
  use adj_mod
  implicit none

  
  !=========================================
  ! declare module variables
  !=========================================
  integer,private                       :: ntraces
  integer,private                       :: ngx,ngy
  real,private                          :: ogx,dgx,ogy,dgy
  real,allocatable,dimension(:),private :: rx,ry,sx,sy
  

  contains


  !=========================================
  ! initialization subroutine
  !=========================================
  subroutine tomo_straightray_init(ntraces_in,rx_in,ry_in,sx_in,sy_in,ngx_in,ogx_in,dgx_in,ngy_in,ogy_in,dgy_in)
    ! declare variables
    integer             :: ntraces_in
    integer             :: ngx_in,ngy_in
    real                :: ogx_in,dgx_in,ogy_in,dgy_in
    real,dimension(:)   :: rx_in,ry_in,sx_in,sy_in

    ! pass number of traces and allocate
    ntraces = ntraces_in
    allocate(rx(ntraces))
    allocate(ry(ntraces))
    allocate(sx(ntraces))
    allocate(sy(ntraces))

    ! pass coordinates into rest of module
    rx = rx_in
    ry = ry_in
    sx = sx_in
    sy = sy_in

    ! pass model parameters into rest of module
    ngx = ngx_in
    dgx = dgx_in
    ogx = ogx_in
    ngy = ngy_in
    dgy = dgy_in
    ogy = ogy_in

  end subroutine


  !=========================================
  ! clean up subroutine
  !=========================================
  subroutine tomo_straightray_close()
    ! deallocate module arrays
    deallocate(rx,ry,sx,sy)

  end subroutine


  !=========================================
  ! operator for SEP solvers
  !=========================================
  function tomo_straightray_lop(adj,add,m,d) result(stat)
    ! declare variables
    integer            :: stat
    logical,intent(in) :: adj,add
    real,dimension(:)  :: m,d

    ! begin calling functions
    call adjnull (adj,add,m,d)
    if (.not.adj) call tomo_straightray_for(m,d)
    if (     adj) call tomo_straightray_adj(m,d)
    stat = 0
 
  end function


  !=========================================
  ! forward tomography subroutine
  !=========================================
  subroutine tomo_straightray_for(m,d)
    ! declare variables
    integer                       :: ii,jj
    integer                       :: irx,iry,isx,isy
    integer                       :: prev_ix,prev_iy,next_ix,next_iy
    integer                       :: xstep,ystep
    real                          :: xoffset,yoffset
    real                          :: xmove1,ymove1,xmove2,ymove2
    real                          :: prev_xcoord,prev_ycoord,next_xcoord,next_ycoord
    real                          :: dist
    real,dimension(:),intent(in)  :: m
    real,dimension(:),intent(out) :: d

    ! loop over all input data to construct G matrix
    do ii = 1,ntraces
      ! determine where the source and receiver are in model domain indices
      irx = 1+int((rx(ii)-ogx)/dgx)
      iry = 1+int((ry(ii)-ogy)/dgy)
      isx = 1+int((sx(ii)-ogx)/dgx)
      isy = 1+int((sy(ii)-ogy)/dgy)

      ! determine which direction to step in (from source to receiver)
      ! indicate which general directions to move in
      if (sx(ii).lt.rx(ii)) then
        xstep = 1
      elseif (sx(ii).gt.rx(ii)) then
        xstep = -1
      else
        xstep = 0
      end if
      if (sy(ii).lt.ry(ii)) then
        ystep = 1
      elseif (sy(ii).gt.ry(ii)) then
        ystep = -1
      else
        ystep = 0
      end if

      ! take care of situation where rays are along boundaries
      if (rx(ii).eq.(ogx+(irx-1)*dgx) .and. xstep.eq. 1) irx = irx-1
      if (ry(ii).eq.(ogy+(iry-1)*dgy) .and. ystep.eq. 1) iry = iry-1
      if (sx(ii).eq.(ogx+(isx-1)*dgx) .and. xstep.eq.-1) isx = isx-1
      if (sy(ii).eq.(ogy+(isy-1)*dgy) .and. ystep.eq.-1) isy = isy-1

      ! break out of loop if indices are out of range
      if (irx.lt.1 .or. irx.gt.ngx) cycle
      if (isx.lt.1 .or. isx.gt.ngx) cycle
      if (iry.lt.1 .or. iry.gt.ngy) cycle
      if (isy.lt.1 .or. isy.gt.ngy) cycle

      ! set initial x and y indices and coordinates
      prev_ix = isx
      prev_iy = isy
      prev_xcoord = sx(ii)
      prev_ycoord = sy(ii)

      ! loop until receiver indices are reached
      do while (prev_ix.ne.irx .or. prev_iy.ne.iry)

        ! calculate total x and y offsets (will effectively provide a slope)
        xoffset = rx(ii)-prev_xcoord
        yoffset = ry(ii)-prev_ycoord

        ! input lengths of potential movements
        ! if we move along x, then compute corresponding y
        if (xstep.eq.1) then
          xmove1 = (ogx+(prev_ix*dgx))-prev_xcoord
          ymove1 = xmove1*(yoffset/(xoffset))
        else if (xstep.eq.-1) then
          xmove1 = (ogx+(prev_ix-1)*dgx)-prev_xcoord
          ymove1 = xmove1*(yoffset/(xoffset))
        else
          xmove1 = 0
          ymove1 = 1/0.
        end if
        ! if we move along y, then compute corresponding x
        if (ystep.eq.1) then
          ymove2 = (ogy+(prev_iy*dgy))-prev_ycoord
          xmove2 = ymove2*(xoffset/(yoffset))
        else if (ystep.eq.-1) then
          ymove2 = (ogy+(prev_iy-1)*dgy)-prev_ycoord
          xmove2 = ymove2*(xoffset/(yoffset))
        else
          ymove2 = 0
          xmove2 = 1/0.
        end if

        ! determine which direction to walk in
        ! step in the +/- x direction
        if (abs(xmove2).gt.abs(xmove1) .and. abs(ymove2).gt.abs(ymove1)) then
          ! update both step index and actual coordinate
          next_ix = prev_ix+xstep
          next_iy = prev_iy
          if (xstep.eq. 1) next_xcoord = ogx+(next_ix-1)*dgx
          if (xstep.eq.-1) next_xcoord = ogx+(next_ix  )*dgx
          next_ycoord = prev_ycoord+(next_xcoord-prev_xcoord)*(yoffset/(xoffset))
        ! step in the +/- y direction
        else if (abs(xmove1).gt.abs(xmove2) .and. abs(ymove1).gt.abs(ymove2)) then
          ! update both step index and actual coordinate
          next_iy = prev_iy+ystep
          next_ix = prev_ix
          if (ystep.eq. 1) next_ycoord = ogy+(next_iy-1)*dgy
          if (ystep.eq.-1) next_ycoord = ogy+(next_iy  )*dgy
          next_xcoord = prev_xcoord+(next_ycoord-prev_ycoord)*(xoffset/(yoffset))
        ! step in the +/- x and y directions
        else
          ! update both step index and actual coordinate
          ! scenario where moving perfectly horizontally or vertically
          next_ix = prev_ix
          next_iy = prev_iy
          next_xcoord = prev_xcoord
          next_ycoord = prev_ycoord
          ! scenario where moving diagonally through model domain nodes
          if (prev_ix.ne.irx) then
            next_ix = prev_ix+xstep
            if (xstep.eq. 1) next_xcoord = ogx+((next_ix-1)*dgx)
            if (xstep.eq.-1) next_xcoord = ogx+(next_ix*dgx)
          end if
          if (prev_iy.ne.iry) then
            next_iy = prev_iy+ystep
            if (ystep.eq. 1) next_ycoord = ogy+((next_iy-1)*dgy)
            if (ystep.eq.-1) next_ycoord = ogy+(next_iy*dgy)
          end if
        end if
        ! calculate length in cell and place in appropriate model index
        dist = sqrt((next_xcoord-prev_xcoord)**2 + (next_ycoord-prev_ycoord)**2)
        jj = ((prev_iy-1)*ngx)+prev_ix
        ! integrate / matrix multiply
        d(ii) = d(ii)+(dist*m(jj))

        ! cycle through coordinates and indices
        prev_xcoord = next_xcoord
        prev_ycoord = next_ycoord
        prev_ix = next_ix
        prev_iy = next_iy

      end do

      ! once receiver cell reached, finish the last bit within that cell
      !dist = sqrt((next_xcoord-prev_xcoord)**2 + (next_ycoord-prev_ycoord)**2)
      dist = sqrt((rx(ii)-prev_xcoord)**2 + (ry(ii)-prev_ycoord)**2)
      jj = ((prev_iy-1)*ngx)+prev_ix
      ! integrate / matrix multiply
      d(ii) = d(ii)+(dist*m(jj))

    end do

  end subroutine


  !=========================================
  ! adjoint tomography subroutine
  !=========================================
  subroutine tomo_straightray_adj(m,d)
    ! declare variables
    integer                       :: ii,jj
    integer                       :: irx,iry,isx,isy
    integer                       :: prev_ix,prev_iy,next_ix,next_iy
    integer                       :: xstep,ystep
    real                          :: xoffset,yoffset
    real                          :: xmove1,ymove1,xmove2,ymove2
    real                          :: prev_xcoord,prev_ycoord,next_xcoord,next_ycoord
    real                          :: dist
    real,dimension(:),intent(out) :: m
    real,dimension(:),intent(in)  :: d

    ! loop over all input data to construct G matrix
    do ii = 1,ntraces
      ! determine where the source and receiver are in model domain indices
      irx = 1+int((rx(ii)-ogx)/dgx)
      iry = 1+int((ry(ii)-ogy)/dgy)
      isx = 1+int((sx(ii)-ogx)/dgx)
      isy = 1+int((sy(ii)-ogy)/dgy)

      ! determine which direction to step in (from source to receiver)
      ! indicate which general directions to move in
      if (sx(ii).lt.rx(ii)) then
        xstep = 1
      elseif (sx(ii).gt.rx(ii)) then
        xstep = -1
      else
        xstep = 0
      end if
      if (sy(ii).lt.ry(ii)) then
        ystep = 1
      elseif (sy(ii).gt.ry(ii)) then
        ystep = -1
      else
        ystep = 0
      end if

      ! take care of situation where rays are along boundaries
      if (rx(ii).eq.(ogx+(irx-1)*dgx) .and. xstep.eq. 1) irx = irx-1
      if (ry(ii).eq.(ogy+(iry-1)*dgy) .and. ystep.eq. 1) iry = iry-1
      if (sx(ii).eq.(ogx+(isx-1)*dgx) .and. xstep.eq.-1) isx = isx-1
      if (sy(ii).eq.(ogy+(isy-1)*dgy) .and. ystep.eq.-1) isy = isy-1

      ! break out of loop if indices are out of range
      if (irx.lt.1 .or. irx.gt.ngx) cycle
      if (isx.lt.1 .or. isx.gt.ngx) cycle
      if (iry.lt.1 .or. iry.gt.ngy) cycle
      if (isy.lt.1 .or. isy.gt.ngy) cycle

      ! set initial x and y indices and coordinates
      prev_ix = isx
      prev_iy = isy
      prev_xcoord = sx(ii)
      prev_ycoord = sy(ii)

      ! loop until receiver indices are reached
      do while (prev_ix.ne.irx .or. prev_iy.ne.iry)

        ! calculate total x and y offsets (will effectively provide a slope)
        xoffset = rx(ii)-prev_xcoord
        yoffset = ry(ii)-prev_ycoord

        ! input lengths of potential movements
        ! if we move along x, then compute corresponding y
        if (xstep.eq.1) then
          xmove1 = (ogx+(prev_ix*dgx))-prev_xcoord
          ymove1 = xmove1*(yoffset/(xoffset))
        else if (xstep.eq.-1) then
          xmove1 = (ogx+(prev_ix-1)*dgx)-prev_xcoord
          ymove1 = xmove1*(yoffset/(xoffset))
        else
          xmove1 = 0
          ymove1 = 1/0.
        end if

        ! if we move along y, then compute corresponding x
        if (ystep.eq.1) then
          ymove2 = (ogy+(prev_iy*dgy))-prev_ycoord
          xmove2 = ymove2*(xoffset/(yoffset))
        else if (ystep.eq.-1) then
          ymove2 = (ogy+(prev_iy-1)*dgy)-prev_ycoord
          xmove2 = ymove2*(xoffset/(yoffset))
        else
          ymove2 = 0
          xmove2 = 1/0.
        end if

        ! determine which direction to walk in
        ! step in the +/- x direction
        if (abs(xmove2).gt.abs(xmove1) .and. abs(ymove2).gt.abs(ymove1)) then
          ! update both step index and actual coordinate
          next_ix = prev_ix+xstep
          next_iy = prev_iy
          if (xstep.eq. 1) next_xcoord = ogx+(next_ix-1)*dgx
          if (xstep.eq.-1) next_xcoord = ogx+(next_ix  )*dgx
          next_ycoord = prev_ycoord+(next_xcoord-prev_xcoord)*(yoffset/(xoffset))
        ! step in the +/- y direction
        else if (abs(xmove1).gt.abs(xmove2) .and. abs(ymove1).gt.abs(ymove2)) then
          ! update both step index and actual coordinate
          next_iy = prev_iy+ystep
          next_ix = prev_ix
          if (ystep.eq. 1) next_ycoord = ogy+(next_iy-1)*dgy
          if (ystep.eq.-1) next_ycoord = ogy+(next_iy  )*dgy
          next_xcoord = prev_xcoord+(next_ycoord-prev_ycoord)*(xoffset/(yoffset))
        ! step in the +/- x and y directions
        else
          ! update both step index and actual coordinate
          ! scenario where moving perfectly horizontally or vertically
          next_ix = prev_ix
          next_iy = prev_iy
          next_xcoord = prev_xcoord
          next_ycoord = prev_ycoord
          ! scenario where moving diagonally through model domain nodes
          if (prev_ix.ne.irx) then
            next_ix = prev_ix+xstep
            if (xstep.eq. 1) next_xcoord = ogx+((next_ix-1)*dgx)
            if (xstep.eq.-1) next_xcoord = ogx+(next_ix*dgx)
          end if
          if (prev_iy.ne.iry) then
            next_iy = prev_iy+ystep
            if (ystep.eq. 1) next_ycoord = ogy+((next_iy-1)*dgy)
            if (ystep.eq.-1) next_ycoord = ogy+(next_iy*dgy)
          end if
        end if

        ! calculate length in cell and place in appropriate model index
        dist = sqrt((next_xcoord-prev_xcoord)**2 + (next_ycoord-prev_ycoord)**2)
        jj = ((prev_iy-1)*ngx)+prev_ix
        ! integrate / matrix multiply
        m(jj) = m(jj)+(dist*d(ii))

        ! cycle through coordinates and indices
        prev_xcoord = next_xcoord
        prev_ycoord = next_ycoord
        prev_ix = next_ix
        prev_iy = next_iy

      end do

      !dist = sqrt((next_xcoord-prev_xcoord)**2 + (next_ycoord-prev_ycoord)**2)
      dist = sqrt((rx(ii)-prev_xcoord)**2 + (ry(ii)-prev_ycoord)**2)
      jj = ((prev_iy-1)*ngx)+prev_ix
      ! integrate / matrix multiply
      m(jj) = m(jj)+(dist*d(ii))

    end do

  end subroutine

end module
